/* Terraform - (C) 1997-2000 Robert Gasch (r.gasch@chello.nl)
 *  - http://terraform.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <math.h>
#include <gnome.h>
#include <libgnome/gnome-help.h>
#include <libgnomeprint/gnome-print.h>
#include <libgnomeprint/gnome-print-master.h>
#include <libgnomeprint/gnome-printer-dialog.h>
#include <libgnomeprint/gnome-print-preview.h>

#include "callbacks.h"
#include "interface.h"
#include "support.h"
#include "support2.h"
#include "tterrain.h"
#include "tundo.h"
#include "mainwindow.h"
#include "terrainwindow.h"
#include "genperlin.h"
#include "genspectral.h"
#include "gensubdiv.h"
#include "genrandom.h"
#include "trgb.h"
#include "tworld.h"
#include "smartexec.h"
#include "filenameops.h"
#include "writer.h"
#include "reader.h"
#include "filelist.h"
#include "filters.h"
#include "filterwindow.h"
#include "place.h"
#include "selection.h"

#define CLIP(x,m,M) MIN (MAX (x, m), M)

typedef struct
{
  gchar     *description;
  gchar     *pattern;
  TFileType  type;
} FileTypeInfo;

FileTypeInfo save_as_file_type_info[] =
{
  { "Automatic (*)", "*", FILE_AUTO },
  { "", "", FILE_UNKNOWN },
  { "Terraform (*.terraform)", "*.terraform", FILE_NATIVE },
  { "Bitmap (*.bmp)", "*.bmp", FILE_BMP },
  { "Targa (*.tga)", "*.tga", FILE_TGA },
  { "Pgm Binary (*.pgm)", "*.pgm", FILE_PGM },
  { "Pg8 Ascii (*.pg8)", "*.pg8", FILE_PG8 },
  { "Matlab (*.mat)", "*.mat", FILE_MAT },
  { "Oct (*.mat)", "*.mat", FILE_MAT },
  { "Autocad (*.ac)", "*.ac", FILE_AC },
  { "VRML (*.vrml)", "*.vrml", FILE_VRML },
  { NULL, NULL, FILE_UNKNOWN }
};

FileTypeInfo open_file_type_info[] =
{
  { "Automatic (*)", "*", FILE_AUTO },
  { "", "", FILE_UNKNOWN },
  { "Terraform (*.terraform)", "*.terraform", FILE_NATIVE },
  { "Bitmap (*.bmp)", "*.bmp", FILE_BMP },
  { "Targa (*.tga)", "*.tga", FILE_TGA },
  { "Gtopo (*.DEM)", "*.DEM", FILE_GTOPO },
  { "Matlab (*.mat)", "*.mat", FILE_MAT },
  { "Oct (*.oct)", "*.oct", FILE_OCT },
  { "Pgm Binary (*.pgm)", "*.pgm", FILE_PGM },
  { "Pg8 Ascii (*.pg8)", "*.pg8", FILE_PG8 },

#ifdef HAVE_GDKPIXBUF
  { "Jpeg (*.jpg)", "*.jpg", FILE_JPG },
  { "Network Graphics (*.png)", "*.png", FILE_PNG },
  { "RAS (*.ras)", "*.ras", FILE_RAS },
  { "Tiff (*.tif)", "*.tif", FILE_TIF },
  { "X Bitmap (*.xbm)", "*.xbm", FILE_XBM },
  { "X Pixmap (*.xpm)", "*.xpm", FILE_XPM },
#endif

  { NULL, NULL, FILE_UNKNOWN }
};


/*
 * create_file_options_menu:
 *   Create a GtkMenu widget given a FileTypeInfo array.
 */

static GtkWidget *
create_file_options_menu (FileTypeInfo *info)
{
  GtkWidget *menu;
  GtkWidget *item;
  int        i;

  menu = gtk_menu_new ();

  for (i = 0; info[i].description != NULL; i++)
    {
      if (info[i].description[0] == '\0')
        item = gtk_menu_item_new ();
      else
        item = gtk_menu_item_new_with_label (info[i].description);

      gtk_container_add (GTK_CONTAINER (menu), item);
      gtk_widget_show (item);
    }

  return menu;
}


static void
on_file_type_selected (GtkWidget *menu,
                       gpointer   user_data)
{
  GtkWidget    *file_sel;
  gchar        *pattern;
  FileTypeInfo *info;

  file_sel = lookup_toplevel (menu);
  info = gtk_object_get_data (GTK_OBJECT (file_sel), "data_file_type_info");
  pattern = info[get_option (file_sel, "type_menu")].pattern;

  gtk_file_selection_complete (GTK_FILE_SELECTION (file_sel), pattern);
}


static void
add_file_type_menu (GtkWidget    *filesel,
                    FileTypeInfo *info)
{
  GtkWidget *frame;
  GtkWidget *file_options_window;
  GtkWidget *file_options_menu;
  GtkWidget *menu;

  frame = gtk_frame_new (NULL);
  gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_NONE);

  /* Cannibalize the "file_options" object from the file_option_window... */
  file_options_window = create_file_options_window ();
  gtk_widget_reparent (lookup_widget (file_options_window, "file_options"),
                       frame);
  file_options_menu = lookup_widget (file_options_window, "type_menu");

  menu = create_file_options_menu (info);
  gtk_signal_connect (GTK_OBJECT (menu),
                      "deactivate",
                      GTK_SIGNAL_FUNC (on_file_type_selected),
                      NULL);

  gtk_option_menu_set_menu (GTK_OPTION_MENU (file_options_menu), menu);
  gtk_widget_destroy (file_options_window);

  /* ... and pack it into the file selection dialog box */
  gtk_box_pack_end (GTK_BOX (GTK_FILE_SELECTION (filesel)->main_vbox),
                    frame, FALSE, FALSE, 5);

  gtk_object_set_data (GTK_OBJECT (filesel), "type_menu", file_options_menu);
  gtk_object_set_data (GTK_OBJECT (filesel), "data_file_type_info", info);

  gtk_widget_show_all (frame);
}


void
on_main_file_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_main_new_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


static gchar *perlin_first[] =
{
  "first", "first_frequency", "first_amplitude",
  "first_iterations", "first_filter"
};

static gchar *perlin_second[] =
{
  "second", "second_frequency", "second_amplitude",
  "second_iterations", "second_filter"
};

static gchar *perlin_third[] =
{
  "third", "third_frequency", "third_amplitude",
  "third_iterations", "third_filter"
};

static gchar *perlin_fourth[] =
{
  "fourth", "fourth_frequency", "fourth_amplitude",
  "fourth_iterations", "fourth_filter"
};

static gchar **perlin_rows[] =
{
  perlin_first, perlin_second, perlin_third, perlin_fourth
};

void
on_main_perlin_noise_activate          (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *perlin_noise;
  GtkWidget *main_window;

  perlin_noise = create_perlin_noise_window ();
  main_window = lookup_toplevel (GTK_WIDGET (menuitem));
  gtk_object_set_data (GTK_OBJECT (perlin_noise), "data_parent", main_window);
  gtk_widget_show (perlin_noise);
}

void
on_perlin_toggled                      (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  gint i, j;

  for (i = 0; i < 4; i++)
    {
      gboolean on;

      on = get_boolean (GTK_WIDGET (togglebutton), perlin_rows[i][0]);

      for (j = 1; j < 5; j++)
        gtk_widget_set_sensitive (lookup_widget (GTK_WIDGET (togglebutton),
                                  perlin_rows[i][j]), on);
    }
}


void
on_main_spectral_activate              (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *spectral, *main_window;

  spectral = create_spectral_window ();
  main_window = lookup_toplevel (GTK_WIDGET (menuitem));
  gtk_object_set_data (GTK_OBJECT (spectral), "data_parent", main_window);
  gtk_widget_show (spectral);
}


void
on_main_subdivision_activate           (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *subdivision, *main_window;

  subdivision = create_subdivision_window ();
  main_window = lookup_toplevel (GTK_WIDGET (menuitem));
  gtk_object_set_data (GTK_OBJECT (subdivision), "data_parent", main_window);
  gtk_widget_show (subdivision);
}


void
on_main_random_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  TTerrain  *terrain;
  GtkWidget *terrain_window;
  GtkWidget *main_window;

  main_window = lookup_toplevel (GTK_WIDGET (menuitem));

  terrain = t_terrain_generate_random ();
  terrain_window = terrain_window_new (main_window, terrain);
  terrain_window_save_state (terrain_window, _("Generate Random"));
  t_terrain_unref (terrain);
}


void
on_main_open_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *open_file;
  GtkWidget *main_window;

  main_window = lookup_toplevel (GTK_WIDGET (menuitem));
  open_file = create_open_file ();
  add_file_type_menu (open_file, open_file_type_info);
  gtk_object_set_data (GTK_OBJECT (open_file), "data_parent", main_window);
  gtk_widget_show (open_file);
}


static void
merge_apply                            (GtkWidget       *merge_window,
                                        gboolean         preview)
{
  TTerrain  *terrain_1;
  TTerrain  *terrain_2;
  TTerrain  *terrain;
  GtkWidget *main_window;
  gint       operator;
  gfloat     weight_1;
  gfloat     weight_2;

  if (preview)
    {
      if (get_boolean (merge_window, "use_preview") == FALSE)
        return;

      terrain_1 = gtk_object_get_data (GTK_OBJECT (merge_window),
                                       "data_top_thumbnail");
      terrain_2 = gtk_object_get_data (GTK_OBJECT (merge_window),
                                       "data_bottom_thumbnail");
    }
  else
    {
      terrain_1 = gtk_object_get_data (GTK_OBJECT (merge_window),
                                       "data_top_t_terrain");
      terrain_2 = gtk_object_get_data (GTK_OBJECT (merge_window),
                                       "data_bottom_t_terrain");
    }

  operator = get_option (merge_window, "merge_op");
  weight_1 = get_float (merge_window, "top_amount");
  weight_2 = get_float (merge_window, "bottom_amount");

  terrain =
    t_terrain_merge (terrain_1, terrain_2, weight_1, weight_2, operator);
  t_terrain_normalize (terrain, FALSE);

  if (preview)
    {
      TRgb     *rgb;
      TTerrain *terrain_old;

      terrain_old = gtk_object_get_data (GTK_OBJECT (merge_window),
                                         "data_t_terrain");
      rgb = gtk_object_get_data (GTK_OBJECT (merge_window), "data_t_rgb");
      if (terrain_old == NULL)
        {
          gtk_object_set_data_full (GTK_OBJECT (merge_window),
                                    "data_t_terrain", terrain,
                                    (GtkDestroyNotify) t_terrain_unref);
        }
      else
        {
          t_terrain_set_size (terrain_old, terrain->width, terrain->height);
          t_terrain_copy_heightfield (terrain, terrain_old);
          t_terrain_unref (terrain);
          terrain = terrain_old;
        }

      terrain->wireframe_resolution = 5;
      t_terrain_paint (terrain, rgb);
      t_rgb_hard_sync (rgb);
    }
  else
    {
      GtkWidget *terrain_window;

      main_window = lookup_widget (merge_window, "data_parent");
      terrain_window = terrain_window_new (main_window, terrain);
      terrain_window_save_state (terrain_window, _("Merge"));
    }
}

void
merge_apply_preview                    (GtkWidget       *merge_window,
                                        TTerrain        *terrain)
{
  merge_apply (merge_window, TRUE);
}


static void
on_merge_adjustment_changed            (GtkAdjustment   *adjustment,
                                        gpointer         merge_window)
{
  merge_apply (merge_window, TRUE);
}


void
on_main_merge_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  TRgb          *rgb;
  TTerrain      *terrain;
  GtkAdjustment *adjustment;
  GtkWidget     *terrain_window;
  GtkWidget     *merge_window;
  GtkWidget     *main_window;
  GtkWidget     *area;
  GtkWidget     *option_menu;
  GtkWidget     *frame;
  gchar         *text;

  main_window = lookup_toplevel (GTK_WIDGET (menuitem));

  merge_window = create_merge_window ();

  option_menu = lookup_widget (merge_window, "merge_op");
  gtk_signal_connect (GTK_OBJECT (GTK_OPTION_MENU (option_menu)->menu),
                      "deactivate",
                      GTK_SIGNAL_FUNC (on_merge_adjustment_changed),
                      merge_window);

  gtk_object_set_data (GTK_OBJECT (merge_window), "data_filter_apply_func",
                       &merge_apply_preview);

  gtk_object_set_data (GTK_OBJECT (merge_window), "data_parent", main_window);

  area = lookup_widget (merge_window, "top_thumbnail");
  terrain_window = main_window_get_terrain_window (main_window, 0);
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");
  t_terrain_ref (terrain);
  gtk_object_set_data_full (GTK_OBJECT (merge_window),
                            "data_top_t_terrain", terrain,
                            (GtkDestroyNotify) t_terrain_unref);
  terrain = t_terrain_clone_preview (terrain);
  gtk_object_set_data_full (GTK_OBJECT (merge_window),
                            "data_top_thumbnail", terrain,
                            (GtkDestroyNotify) t_terrain_unref);
  rgb = t_rgb_new (area, terrain->width, terrain->height);
  gtk_object_set_data_full (GTK_OBJECT (merge_window),
                            "data_top_t_rgb", rgb,
                            (GtkDestroyNotify) t_rgb_free);
  t_terrain_paint_preview (terrain, rgb);
  frame = lookup_widget (merge_window, "top_frame");
  text = terrain_window_get_title (terrain_window);
  gtk_frame_set_label (GTK_FRAME (frame), text);
  g_free (text);

  area = lookup_widget (merge_window, "bottom_thumbnail");
  terrain_window = main_window_get_terrain_window (main_window, 1);
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");
  t_terrain_ref (terrain);
  gtk_object_set_data_full (GTK_OBJECT (merge_window),
                            "data_bottom_t_terrain", terrain,
                            (GtkDestroyNotify) t_terrain_unref);
  terrain = t_terrain_clone_preview (terrain);
  gtk_object_set_data_full (GTK_OBJECT (merge_window),
                            "data_bottom_thumbnail", terrain,
                            (GtkDestroyNotify) t_terrain_unref);
  rgb = t_rgb_new (area, terrain->width, terrain->height);
  gtk_object_set_data_full (GTK_OBJECT (merge_window),
                            "data_bottom_t_rgb", rgb,
                            (GtkDestroyNotify) t_rgb_free);
  t_terrain_paint_preview (terrain, rgb);
  frame = lookup_widget (merge_window, "bottom_frame");
  text = terrain_window_get_title (terrain_window);
  gtk_frame_set_label (GTK_FRAME (frame), text);
  g_free (text);

  adjustment =
    GTK_RANGE (lookup_widget (merge_window, "top_amount"))->adjustment;
  gtk_signal_connect (GTK_OBJECT (adjustment), "value_changed",
                      (GtkSignalFunc) &on_merge_adjustment_changed,
                      merge_window);

  adjustment =
    GTK_RANGE (lookup_widget (merge_window, "bottom_amount"))->adjustment;
  gtk_signal_connect (GTK_OBJECT (adjustment), "value_changed",
                      (GtkSignalFunc) &on_merge_adjustment_changed,
                      merge_window);


  area = lookup_widget (merge_window, "preview");
  rgb = t_rgb_new (area, terrain->width, terrain->height);
  gtk_object_set_data_full (GTK_OBJECT (merge_window),
                            "data_t_rgb", rgb, (GtkDestroyNotify) t_rgb_free);

  main_window_copy_objects (main_window,
    GTK_CLIST (lookup_widget (merge_window, "merge_list")));

  merge_apply (merge_window, TRUE);
  gtk_widget_show (merge_window);
}


void
gamma_preview_update (GtkWidget *option_window)
{
  TRgb   *rgb;
  gfloat  gamma;
  gint    value;

  rgb = gtk_object_get_data (GTK_OBJECT (option_window), "data_t_rgb");
  gamma = get_float (option_window, "gamma");

  value = rint (pow (0.5, 1.0 / gamma) * 255.0);
  t_rgb_set_fore (rgb, value, value, value);
  t_rgb_fill_rect (rgb, 0, 0, 49, 49);
  t_rgb_hard_sync (rgb);
}


void
gamma_preview_initialize (GtkWidget *option_window)
{
  TRgb *rgb;
  gint  x, y;

  rgb = gtk_object_get_data (GTK_OBJECT (option_window), "data_t_rgb");
  for (y = 0; y < 50; y++)
    for (x = 50; x < 100; x++)
      if (((x >> 1) + (y >> 1)) & 1)
        t_rgb_draw_pel (rgb, x, y);

  gamma_preview_update (option_window);
}


void
on_main_options_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  TRgb      *rgb;
  GtkWidget *main_window;
  GtkWidget *gamma_preview;
  GtkWidget *options;
  GtkWidget *gamma;

  main_window = lookup_toplevel (GTK_WIDGET (menuitem));
  options = create_options_window ();
  gamma_preview = lookup_widget (options, "gamma_preview");
  rgb = t_rgb_new (gamma_preview, 100, 50);
  gtk_object_set_data_full (GTK_OBJECT (options),
                            "data_t_rgb", rgb, (GtkDestroyNotify) t_rgb_free);
  gamma_preview_initialize (options);
  gamma = lookup_widget (options, "gamma");
  gtk_signal_connect_object (
            GTK_OBJECT (GTK_RANGE (gamma)->adjustment),
            "value_changed",
            GTK_SIGNAL_FUNC (&gamma_preview_update),
            GTK_OBJECT (options));

  gtk_object_set_data (GTK_OBJECT (options), "data_parent", main_window);
  main_window_state_unpack_options (main_window, options);
  gtk_widget_show (options);
}


void
on_main_print_settings_activate        (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
}


void
on_main_exit_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  main_window_exit_sequence (lookup_toplevel (GTK_WIDGET (menuitem)));
}


void
on_main_help_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_main_about_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *about_window = create_about_window ();

  gtk_widget_show (about_window);
}


void
on_terrain_save_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *terrain_window;
  GtkWidget *main_window;
  TTerrain  *terrain;
  gint       status;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  main_window = gtk_object_get_data (GTK_OBJECT (terrain_window),
                                     "data_parent");
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");
  status = t_terrain_save (terrain, FILE_NATIVE, NULL);
  terrain_window_set_title (terrain_window, terrain_window);
  main_window_update_terrain_window (main_window, terrain_window);

  if (status != 0)
    {
      GtkWidget *error_window;

      error_window = create_error_window ();
      set_text (error_window, "error_label", _("An error occurred while saving the file."));
      gtk_widget_show (error_window);
    }
}


void
on_terrain_save_as_activate            (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *save_as;
  GtkWidget *terrain_window;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  save_as = create_save_as ();
  add_file_type_menu (save_as, save_as_file_type_info);
  gtk_object_set_data (GTK_OBJECT (save_as), "data_parent", terrain_window);
  gtk_widget_show (save_as);
}


typedef GList Cleanup;

static void
cleanup (gpointer user_data)
{
  Cleanup *list = user_data;

  list = g_list_first (list);
  while (list != NULL)
    {
      unlink ((gchar*) list->data);
      g_free (list->data);

      list = list->next;
    }

  g_list_free (list);
}


void
on_terrain_gimp_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *terrain_window;
  TTerrain  *terrain;
  gchar     *tmp, *base, *bmp;
  gchar     *command;
  GList     *cleanup_list;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  base = filename_get_base (terrain->filename);
  tmp = g_strdup_printf ("/tmp/%s", base);
  g_free (base);

  bmp = filename_new_extension (tmp, "bmp");
  t_terrain_save (terrain, FILE_BMP, bmp);
  g_free (tmp);

  command = g_strdup_printf ("gimp %s", bmp);

  cleanup_list = NULL;
  cleanup_list = g_list_prepend (cleanup_list, bmp);

  execute_command (command, _("There was an error executing The GIMP"), cleanup,
                   cleanup_list);
  g_free (command);
}


void
on_terrain_open_gl_activate            (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


void
on_terrain_close_activate              (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *terrain_window;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  terrain_window_exit_sequence (terrain_window);
}


void
on_terrain_clone_activate              (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  TTerrain  *terrain;
  GtkWidget *terrain_window;
  GtkWidget *main_window;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window),
                                 "data_t_terrain");
  main_window = gtk_object_get_data (GTK_OBJECT (terrain_window),
                                     "data_parent");

  terrain = t_terrain_clone (terrain);
  g_free (terrain->filename);
  terrain->filename = main_window_filename_new (main_window);
  terrain_window = terrain_window_new (main_window, terrain);
  terrain_window_save_state (terrain_window, _("Clone"));
  t_terrain_unref (terrain);
}


void
undo_apply                             (GtkWidget       *undo_window,
                                        TTerrain        *terrain)
{
  GtkWidget *tree;
  GList     *selection;
  gboolean   apply_preview;

  tree = lookup_widget (undo_window, "tree");
  apply_preview =
    gtk_object_get_data (GTK_OBJECT (undo_window), "data_t_terrain") == terrain;

  selection = GTK_TREE_SELECTION (GTK_TREE (tree));
  if (selection != NULL)
    {
      TUndoState *undo_state;

      undo_state =
        gtk_object_get_data (GTK_OBJECT (selection->data), "data_state");

      if (apply_preview)
        {
          TTerrain *undo_terrain, *preview_terrain;

          undo_terrain = t_terrain_clone (terrain);
          t_undo_state_revert (undo_state, undo_terrain);
          preview_terrain = t_terrain_clone_preview (undo_terrain);
          t_terrain_set_size (terrain, preview_terrain->width,
                              preview_terrain->height);
          t_terrain_copy_heightfield (preview_terrain, terrain);
          t_terrain_unref (preview_terrain);
          t_terrain_unref (undo_terrain);
        }
      else
        {
          TUndo    *undo;
          TTerrain *main_terrain, *terrain_window;

          terrain_window = gtk_object_get_data (GTK_OBJECT (undo_window),
                                             "data_parent");
          undo = gtk_object_get_data (GTK_OBJECT (terrain_window),
                                      "data_t_undo");
          main_terrain = gtk_object_get_data (GTK_OBJECT (terrain_window),
                                              "data_t_terrain");

          if (terrain == main_terrain)
            t_undo_revert_state (undo, undo_state, terrain);
          else
            t_undo_state_revert (undo_state, terrain);
        }
    }
}


void
on_terrain_undo_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *terrain_window;
  GtkWidget *modal;
  TUndo     *undo;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  undo = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_undo");

  modal = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_modal");
  if (modal == NULL)
    {
      GtkWidget *undo_window, *tree;
      TRgb      *rgb;
      TTerrain  *terrain;
      TTerrain  *terrain_preview;
      GtkWidget *item;

      gtk_widget_set_sensitive (terrain_window, FALSE);
      undo_window = create_undo_tree_window ();
      gtk_object_set_data (GTK_OBJECT (undo_window),
                           "data_filter_apply_func", &undo_apply);
      gtk_object_set_data (GTK_OBJECT (undo_window),
                           "show_black", &undo_window);

      terrain = gtk_object_get_data (GTK_OBJECT (terrain_window),
                                     "data_t_terrain");
      gtk_object_set_data (GTK_OBJECT (terrain_window), "data_modal",
                           undo_window);
      gtk_object_set_data_full (GTK_OBJECT (undo_window), "data_parent",
                                terrain_window, terrain_window_modal_destroy);
      terrain_preview = t_terrain_clone_preview (terrain);
      gtk_object_set_data_full (GTK_OBJECT (undo_window), "data_t_terrain",
                                terrain_preview,
                                (GtkDestroyNotify) t_terrain_unref);
      rgb = t_rgb_new (lookup_widget (undo_window, "preview"),
                       terrain_preview->width, terrain_preview->height);
      gtk_object_set_data_full (GTK_OBJECT (undo_window), "data_t_rgb",
                                rgb, (GtkDestroyNotify) t_rgb_free);
      terrain_window_set_title (terrain_window, undo_window);

      tree = lookup_widget (undo_window, "tree");
      item = t_undo_unpack_tree (undo, GTK_TREE (tree));
      filter_window_update_preview (undo_window);
      gtk_widget_show (undo_window);
      gtk_tree_select_child (GTK_TREE (tree), item);
    }
}


static void
level_connector_apply                  (GtkWidget       *window,
                                        TTerrain        *terrain)
{
  gint count;

  count = get_int (window, "count");
  t_terrain_connect (terrain, count);
}


void
on_terrain_connect_activate            (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = { NULL };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_level_connector_window,
                     &level_connector_apply,
                     hscales,
                     TRUE);
}


static void
craters_apply                          (GtkWidget       *crater_window,
                                        TTerrain        *terrain)
{
  gint       count;
  gfloat     radius, height, coverage, center_x, center_y;
  gboolean   wrap;
  gint       seed;

  center_x = get_float (crater_window, "center_x");
  center_y = get_float (crater_window, "center_y");
  count = get_int (crater_window, "count");
  radius = get_float (crater_window, "radius");
  height = get_float (crater_window, "height");
  coverage = get_float (crater_window, "coverage");
  wrap = get_boolean (crater_window, "wrap");
  if (get_boolean (crater_window, "new_seed"))
    seed = -1;
  else
    seed = get_int (crater_window, "seed");

  gtk_widget_set_sensitive (lookup_widget (crater_window, "center_x"),
                            count == 1);
  gtk_widget_set_sensitive (lookup_widget (crater_window, "center_x_label"),
                            count == 1);
  gtk_widget_set_sensitive (lookup_widget (crater_window, "center_y"),
                            count == 1);
  gtk_widget_set_sensitive (lookup_widget (crater_window, "center_y_label"),
                            count == 1);

  t_terrain_craters (terrain, count, wrap, height, radius, coverage,
                     seed, center_x, center_y);
  t_terrain_normalize (terrain, TRUE);
}


void
on_terrain_craters_activate            (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = {
    "center_x", "center_y", "count", "radius", "height", "coverage", NULL
  };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_craters_window,
                     &craters_apply,
                     hscales,
                     TRUE);
}


static void
fill_apply                             (GtkWidget       *fill_window,
                                        TTerrain        *terrain)
{
  gfloat fill_elevation, fill_tightness;

  fill_elevation = get_float (fill_window, "fill_elevation");
  fill_tightness = get_float (fill_window, "fill_tightness");
  t_terrain_fill (terrain, fill_elevation, fill_tightness);
}


void
on_terrain_fill_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = { "fill_elevation", "fill_tightness", NULL };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_fill_window,
                     &fill_apply,
                     hscales,
                     TRUE);
}


static void
fold_apply (GtkWidget *fold_window,
            TTerrain  *terrain)
{
  gfloat    fold_offset;
  gint      margin;

  fold_offset = get_float (fold_window, "fold_offset");
  margin = fold_offset * MIN (terrain->width, terrain->height);

  t_terrain_fold (terrain, margin);
}


void
on_terrain_fold_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = { "fold_offset", NULL };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_fold_window,
                     &fold_apply,
                     hscales,
                     TRUE);
}


static void
gaussian_hill_apply (GtkWidget     *hill_window,
                     TTerrain      *terrain)
{
  gfloat    center_x, center_y;
  gfloat    radius, scale_factor, radius_factor;
  gfloat    smoothing_factor, delta_scale_factor;

  center_x = get_float (hill_window, "center_x");
  center_y = get_float (hill_window, "center_y");
  radius = get_float (hill_window, "hill_radius");
  scale_factor = get_float (hill_window, "hill_scale_factor");
  radius_factor = get_float (hill_window, "hill_radius_factor");
  smoothing_factor = get_float (hill_window, "hill_smoothing_factor");
  delta_scale_factor = get_float (hill_window, "hill_delta_scale_factor");

  t_terrain_gaussian_hill (terrain, center_x, center_y,
                           radius, radius_factor,
                           scale_factor, smoothing_factor,
                           delta_scale_factor);
  t_terrain_normalize (terrain, TRUE);
}


void
on_terrain_gaussian_hill_activate      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = {
    "center_x", "center_y", "hill_radius", "hill_scale_factor",
    "hill_radius_factor", "hill_smoothing_factor", "hill_delta_scale_factor",
    NULL
  };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_gaussian_hill_window,
                     &gaussian_hill_apply,
                     hscales,
                     TRUE);
}


static void
radial_scale_apply                (GtkWidget *radial_window,
                                   TTerrain  *terrain)
{
  gfloat    center_x, center_y;
  gfloat    start_dist, end_dist;
  gfloat    scale_factor, smooth_factor;

  center_x = get_float (radial_window, "center_x");
  center_y = get_float (radial_window, "center_y");
  start_dist = get_float (radial_window, "scale_start_distance");
  end_dist = get_float (radial_window, "scale_end_distance");
  scale_factor = get_float (radial_window, "scale_factor");
  smooth_factor = get_float (radial_window, "smoothing_factor");
  if (get_boolean (radial_window, "invert"))
    scale_factor = 1.0 / scale_factor;

  t_terrain_radial_scale (terrain, center_x, center_y,
                          scale_factor, start_dist, end_dist, smooth_factor);
  t_terrain_normalize (terrain, TRUE);
}


void
on_terrain_radial_scale_activate       (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = {
    "center_x", "center_y", "scale_start_distance", "scale_end_distance",
    "scale_factor", "smoothing_factor", NULL
  };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_radial_scale_window,
                     &radial_scale_apply,
                     hscales,
                     TRUE);
}


static gchar *mirror_options[] = {
  "mirror_horizontal", "mirror_vertical",
  "mirror_diagonal_1", "mirror_diagonal_2", NULL
};


static void
mirror_apply                           (GtkWidget       *mirror_window,
                                        TTerrain        *terrain)
{
  gint       direction;

  direction = get_radio_menu (mirror_window, mirror_options);
  t_terrain_mirror (terrain, direction);
}


static GtkWidget *
create_mirror_window_advanced          (TTerrain        *terrain)
{
  GtkWidget *mirror_window;

  mirror_window = create_mirror_window ();

  if (terrain->width != terrain->height)
    {
      gtk_widget_set_sensitive (lookup_widget (mirror_window,
                                "mirror_diagonal_1"), FALSE);
      gtk_widget_set_sensitive (lookup_widget (mirror_window,
                                "mirror_diagonal_2"), FALSE);
    }

  return mirror_window;
}


void
on_terrain_mirror_activate             (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = { NULL };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_mirror_window_advanced,
                     &mirror_apply,
                     hscales,
                     TRUE);
}


static void
move_apply                             (GtkWidget       *mirror_window,
                                        TTerrain        *terrain)
{
  gfloat x_offset, y_offset;

  x_offset = get_float (mirror_window, "x_offset");
  y_offset = get_float (mirror_window, "y_offset");

  t_terrain_move (terrain, x_offset, y_offset);
}


void
on_terrain_move_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = { "x_offset", "y_offset", NULL };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_move_window,
                     &move_apply,
                     hscales,
                     TRUE);
}


static gchar *rotate_options[] = {
  "rotate_90", "rotate_180", "rotate_270", NULL
};


void
rotate_apply                           (GtkWidget       *rotate_window,
                                        TTerrain        *terrain)
{
  gint       angle;

  angle = get_radio_menu (rotate_window, rotate_options);
  t_terrain_rotate (terrain, angle);
}


void
on_terrain_rotate_activate             (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = { NULL };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_rotate_window,
                     &rotate_apply,
                     hscales,
                     TRUE);
}


static void
roughen_apply                          (GtkWidget       *roughen_window,
                                        TTerrain        *terrain)
{
  gfloat    factor;
  gboolean  big_grid;

  factor = get_float (roughen_window, "roughen_factor");
  big_grid = get_boolean (roughen_window, "roughen_big");

  t_terrain_roughen_smooth (terrain, TRUE, big_grid, factor);
}


void
on_terrain_roughen_activate            (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = { "roughen_factor", NULL };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_roughen_window,
                     &roughen_apply,
                     hscales,
                     TRUE);
}


static void
smooth_apply                           (GtkWidget       *smooth_window,
                                        TTerrain        *terrain)
{
  gfloat    factor;
  gboolean  big_grid;

  factor = get_float (smooth_window, "smooth_factor");
  big_grid = get_boolean (smooth_window, "smooth_big");

  t_terrain_roughen_smooth (terrain, FALSE, big_grid, factor);
}


void
on_terrain_smooth_activate             (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = { "smooth_factor", NULL };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_smooth_window,
                     &smooth_apply,
                     hscales,
                     TRUE);
}


void
on_smooth_ok_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
}


static void
terrace_apply                          (GtkWidget       *terrace_window,
                                        TTerrain        *terrain)
{
  gint      level_count;
  gfloat    factor;
  gboolean  adjust_sealevel;

  level_count = get_int (terrace_window, "terrace_levels");
  factor = get_float (terrace_window, "terrace_tightness");
  adjust_sealevel = get_boolean (terrace_window, "terrace_adjust_waterlevel");

  t_terrain_terrace (terrain, level_count, factor, adjust_sealevel);
}


void
on_terrain_terrace_activate            (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = { "terrace_levels", "terrace_tightness", NULL };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_terrace_window,
                     &terrace_apply,
                     hscales,
                     TRUE);
}


static void
tile_apply                             (GtkWidget       *tile_window,
                                        TTerrain        *terrain)
{
  TTerrain *preview_terrain;
  gfloat    tile_offset;

  preview_terrain =
    gtk_object_get_data (GTK_OBJECT (tile_window), "data_t_terrain");

  tile_offset = get_float (tile_window, "tile_offset");
  t_terrain_tiler (terrain, tile_offset);

  if (preview_terrain == terrain)
    t_terrain_tile (terrain, 2);
}


void
on_terrain_tile_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = { "tile_offset", NULL };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_tile_window,
                     &tile_apply,
                     hscales,
                     TRUE);
}


static void
transform_apply                        (GtkWidget       *transform_window,
                                        TTerrain        *terrain)
{
  gfloat   sea_threshold, sea_depth, sea_dropoff;
  gfloat   above_power, below_power;

  sea_threshold = get_float (transform_window, "sea_threshold");
  sea_depth = get_float (transform_window, "sea_depth");
  sea_dropoff = get_float (transform_window, "sea_dropoff");
  above_power = get_float (transform_window, "above_power");
  below_power = get_float (transform_window, "below_power");

  if (gtk_object_get_data (GTK_OBJECT (transform_window),
                           "data_t_terrain") == terrain)
    {
      TRgb *rgb;

      rgb = gtk_object_get_data (GTK_OBJECT (transform_window),
                                 "data_function_t_rgb");
      t_terrain_draw_transform (rgb, sea_threshold, sea_depth, sea_dropoff,
                                above_power, below_power);
      t_rgb_hard_sync (rgb);
    }
  terrain->sealevel = sea_threshold;
  t_terrain_transform (terrain, sea_threshold, sea_depth, sea_dropoff,
                                above_power, below_power);
  terrain->sealevel = sea_depth;
}


static GtkWidget *
create_transform_window_advanced       (TTerrain        *terrain)
{
  GtkWidget *widget;
  TRgb      *rgb;

  widget = create_transform_window ();
  rgb = t_rgb_new (lookup_widget (widget, "function"), 250, 102);
  gtk_object_set_data_full (GTK_OBJECT (widget), "data_function_t_rgb", rgb,
                            (GtkDestroyNotify) t_rgb_free);
  set_float (widget, "sea_threshold", terrain->sealevel);
  set_float (widget, "sea_depth", terrain->sealevel);

  return widget;
}


void
on_terrain_transform_activate          (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = {
    "sea_threshold", "sea_depth", "sea_dropoff",
    "above_power", "below_power", NULL
  };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_transform_window_advanced,
                     &transform_apply,
                     hscales,
                     TRUE);
}


void
on_terrain_invert_activate             (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  TTerrain *terrain;
  GtkWidget *terrain_window;
  GtkWidget *main_window;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");
  main_window = gtk_object_get_data (GTK_OBJECT (terrain_window),
                                     "data_parent");

  t_terrain_invert (terrain);
  terrain_window_save_state (terrain_window, _("Invert"));

  terrain->modified = TRUE;
  terrain_window_refresh (terrain_window, FALSE);
}


static void
erode_apply                            (GtkWidget       *terrain_window,
                                        TTerrain        *terrain)
{
  gint      iterations;
  gint      age;
  gfloat    threshold;
  gboolean  trim;
  gchar    *filename;
  gint      frame_count;
  gboolean  single_flow_direction;
  gboolean  observe_sealevel;

  iterations = get_int (terrain_window, "erode_iterations");
  age = get_int (terrain_window, "erode_age");
  threshold = get_float (terrain_window, "erode_threshold");
  trim = get_boolean (terrain_window, "erode_trim");
  if (get_boolean (terrain_window, "erode_save"))
    filename = get_text (terrain_window, "erode_file");
  else
    filename = NULL;
  frame_count = get_int (terrain_window, "erode_frame_count");
  single_flow_direction = get_boolean (terrain_window, "erode_single");
  observe_sealevel = get_boolean (terrain_window, "erode_sealevel");

  t_terrain_erode (terrain, iterations, age, threshold,
                   filename, frame_count, trim,
                   single_flow_direction, !observe_sealevel, NULL);

}


void
on_terrain_erode_activate              (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = { NULL };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_erode_window,
                     &erode_apply,
                     hscales,
                     TRUE);
}


static void
flowmap_apply                          (GtkWidget       *terrain_window,
                                        TTerrain        *terrain)
{
  TTerrain *flowmap;
  gboolean  single_flow_direction;
  gboolean  observe_sealevel;

  single_flow_direction = get_boolean (terrain_window, "flowmap_single");
  observe_sealevel = get_boolean (terrain_window, "flowmap_sealevel");

  flowmap = t_terrain_flowmap (terrain, single_flow_direction,
                               !observe_sealevel, 1.0);

  t_terrain_copy_heightfield (flowmap, terrain);
  t_terrain_free (flowmap);
}


void
on_terrain_flowmap_activate            (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = { NULL };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_flowmap_window,
                     &flowmap_apply,
                     hscales,
                     TRUE);
}


void
on_terrain_scale_activate              (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *scale_window;
  GtkWidget *terrain_window;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  gtk_widget_set_sensitive (terrain_window, FALSE);
  scale_window = create_scale_window ();
  gtk_object_set_data_full (GTK_OBJECT (scale_window), "data_parent",
                            terrain_window, terrain_window_modal_destroy);
  gtk_object_set_data (GTK_OBJECT (terrain_window), "data_modal",
                       scale_window);

  gtk_widget_show (scale_window);
}


void
on_terrain_auto_rotate_activate        (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{

}


gboolean
on_main_window_delete_event            (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  return main_window_exit_sequence (lookup_toplevel (GTK_WIDGET (widget)));
}


GtkWidget*
create_lc_custom (gchar *widget_name, gchar *string1, gchar *string2,
                gint int1, gint int2)
{
  return gtk_label_new ("Foo");
}


void
on_exit_yes_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_destroy (lookup_toplevel (GTK_WIDGET (button)));

  gtk_main_quit ();
}


GtkWidget*
create_render_custom (gchar *widget_name, gchar *string1, gchar *string2,
                gint int1, gint int2)
{
  return gtk_drawing_area_new ();
}


void
on_close_yes_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *close_anyways_window, *terrain_window, *main_window;
  GtkWidget *modal;

  close_anyways_window = lookup_toplevel (GTK_WIDGET (button));
  terrain_window = gtk_object_get_data (GTK_OBJECT (close_anyways_window),
                                        "data_parent");
  main_window = gtk_object_get_data (GTK_OBJECT (terrain_window),
                                     "data_parent");

  modal = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_modal");
  if (modal != NULL)
    gtk_widget_destroy (modal);

  main_window_remove_terrain_window (main_window, terrain_window);

  gtk_widget_destroy (close_anyways_window);
}


gboolean
on_terrain_window_delete_event         (GtkWidget       *widget,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  return terrain_window_exit_sequence (widget);
}


gboolean
on_terrain_button_press_event          (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
  GtkWidget    *terrain_window;
  TTerrain     *terrain;

  terrain_window = lookup_toplevel (widget);
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  if (event->button == 1 && 
      (terrain->view == T_VIEW_2D_PLANE || terrain->view == T_VIEW_3D_WIRE))
    {
      guint *anchor;

      anchor = g_new (guint, 2);
      anchor[0] = event->x;
      anchor[1] = event->y;
      gtk_object_set_data_full (GTK_OBJECT (terrain_window),
                                "data_anchor", anchor, g_free);
      if (terrain->view == T_VIEW_2D_PLANE  &&
          get_boolean (terrain_window, "move"))
        {
          gint    i;
          gfloat  x, y;
          gfloat  limit;
          GArray *array;
          TRgb   *rgb;

          rgb = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_rgb");
          array = terrain->objects;

          x = ((gfloat) event->x) / (rgb->width - 1);
          y = ((gfloat) event->y) / (rgb->height - 1);

          limit = 4.0 / MIN (rgb->width, rgb->height);

          for (i = 0; i < array->len; i++)
            {
              TTerrainObject *object;

              object = &g_array_index (array, TTerrainObject, i);

              if (fabs (x - object->x) < limit &&
                  fabs (y - object->y) < limit)
                {
                  gint *item;

                  item = g_new (gint, 1);
                  *item = i;
                  gtk_object_set_data_full (GTK_OBJECT (terrain_window),
                                            "data_object", item, g_free);
                  break;
                }
            }

        }
      gtk_grab_add (widget);
    }
  else
    {
      gtk_object_remove_data (GTK_OBJECT (terrain_window), "data_anchor");
    }


  return FALSE;
}


void
on_generic_new_seed_toggled            (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  GtkWidget *widget;
  gboolean   value;

  value = !gtk_toggle_button_get_active (togglebutton);

  widget = lookup_widget (GTK_WIDGET (togglebutton), "seed_label");
  gtk_widget_set_sensitive (widget, value);
  widget = lookup_widget (GTK_WIDGET (togglebutton), "seed");
  gtk_widget_set_sensitive (widget, value);
}


void
on_terrain_options_activate            (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *terrain_window;
  GtkWidget *modal;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));

  modal = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_modal");
  if (modal == NULL)
    {
      TRgb      *rgb;
      GtkWidget *drawing_area;
      TTerrain  *terrain;
      TTerrain  *terrain_preview;
      GtkWidget *options_window;
      GtkWidget *theme_menu;
      GList     *list;

      gtk_widget_set_sensitive (terrain_window, FALSE);
      options_window = create_render_options_window ();

      terrain =
        gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

      drawing_area = gtk_object_get_data (GTK_OBJECT (options_window),
                                          "preview");
      terrain_preview = t_terrain_clone_preview (terrain);
      rgb = t_rgb_new (drawing_area, terrain_preview->width,
                       terrain_preview->height);
      gtk_object_set_data_full (GTK_OBJECT (options_window), "data_t_rgb",
                                rgb, (GtkDestroyNotify) &t_rgb_free);
      gtk_object_set_data_full (GTK_OBJECT (options_window), "data_t_terrain",
                                terrain_preview,
                                (GtkDestroyNotify) &t_terrain_unref);

      terrain_preview->view = T_VIEW_2D_PLANE;
      t_terrain_paint (terrain_preview, rgb);

      gtk_object_set_data_full (GTK_OBJECT (options_window),
                                "data_parent", terrain_window,
                                terrain_window_modal_destroy);
      gtk_object_set_data (GTK_OBJECT (terrain_window),
                           "data_modal", options_window);

      list = theme_list_new ();
      theme_menu = gtk_menu_new ();
      if (list == NULL)
        list = g_list_append (list, g_strdup (_("[No Themes Installed]")));

      gtk_object_set_data_full (GTK_OBJECT (options_window),
                                "data_themes", list,
                                (GtkDestroyNotify) file_list_free);

      for (; list != NULL; list = list->next)
        {
          GtkWidget *theme_menuitem;
          guchar    *friendly_name;

          friendly_name = filename_get_friendly ((gchar*) list->data);
          theme_menuitem = gtk_menu_item_new_with_label (friendly_name);
          g_free (friendly_name);

          gtk_widget_show (theme_menuitem);
          gtk_menu_append (GTK_MENU (theme_menu), theme_menuitem);
        }

      gtk_option_menu_set_menu (
        GTK_OPTION_MENU (lookup_widget (options_window, "render_theme")),
        theme_menu);

      /* Now that the theme list is set up, unpack the TTerrain structure. */
      t_terrain_unpack_options (terrain, options_window);

      gtk_widget_show (GTK_WIDGET (options_window));
    }
}


void
terrain_toggle_activate                (GtkMenuShell    *menushell,
                                        gpointer         user_data)
{
  GtkWidget *terrain_window;

  terrain_window = lookup_toplevel (GTK_WIDGET (menushell));
  if (GTK_WIDGET_VISIBLE (terrain_window))
    terrain_window_pack_menu (terrain_window);
}


void
on_load_ok_clicked                     (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget    *open_file;
  GtkWidget    *main_window;
  TTerrain     *terrain;
  gchar        *filename;
  TFileType     type;
  FileTypeInfo *info;

  open_file = lookup_toplevel (GTK_WIDGET (button));
  main_window = gtk_object_get_data (GTK_OBJECT (open_file), "data_parent");
  filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (open_file));
  info = gtk_object_get_data (GTK_OBJECT (open_file), "data_file_type_info");
  type = info[get_option (open_file, "type_menu")].type;
  terrain = t_terrain_load (filename, type);
  if (terrain != NULL)
    {
      GtkWidget *terrain_window;

      terrain_window = terrain_window_new (main_window, terrain);
      terrain_window_save_state (terrain_window, _("Load"));
      t_terrain_unref (terrain);
    }
  else
    {
      GtkWidget *error_window;

      error_window = create_error_window ();
      set_text (error_window, "error_label", _("An error occurred while opening the file."));
      gtk_widget_show (error_window);
    }

  gtk_widget_destroy (open_file);
}


void
on_option_ok_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *options_window;
  GtkWidget *main_window;
  gchar     *file;
  gchar     *home_dir;
  MainState *state;

  options_window = lookup_toplevel (GTK_WIDGET (button));
  main_window = gtk_object_get_data (GTK_OBJECT (options_window), "data_parent");
  main_window_state_pack_options (main_window, options_window);
  state = gtk_object_get_data (GTK_OBJECT (main_window), "data_state");

  home_dir = getenv ("HOME");
  if (home_dir != NULL)
    {
      file = g_strdup_printf ("%s/.terraformrc", home_dir);
      main_window_state_save (state, file);
      g_free (file);
    }

  gtk_widget_destroy (options_window);
}


void
on_option_povray_browse_clicked        (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *find_window;
  GtkWidget *options_window;

  options_window = lookup_toplevel (GTK_WIDGET (button));
  find_window = create_find_povray_window ();
  gtk_object_set_data (GTK_OBJECT (find_window), "data_parent", options_window);
  gtk_widget_show (find_window);
}


void
on_povray_yes_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget   *find_window;
  GtkWidget   *options_window;
  const gchar *povray;

  find_window = lookup_toplevel (GTK_WIDGET (button));
  options_window = gtk_object_get_data (GTK_OBJECT (find_window),
                                        "data_parent");

  povray = povray_probe ();
  if (povray != NULL)
    {
      GtkWidget *text;
      gint       pos;

      text = lookup_widget (options_window, "option_povray");
      gtk_editable_delete_text (GTK_EDITABLE (text), 0, -1);
      pos = 0;
      gtk_editable_insert_text (GTK_EDITABLE (text), povray,
                                strlen (povray), &pos);
    }

  gtk_widget_destroy (find_window);
}


void
on_povray_no_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *find_window;

  find_window = lookup_toplevel (GTK_WIDGET (button));
  gtk_widget_destroy (find_window);
}


void
on_render_ok_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *options_window;
  GtkWidget *terrain_window;
  TTerrain  *terrain;

  options_window = lookup_toplevel (GTK_WIDGET (button));
  terrain_window = gtk_object_get_data (GTK_OBJECT (options_window),
                                        "data_parent");
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  t_terrain_pack_options (terrain, options_window);

  terrain_window_refresh (terrain_window, FALSE);

  gtk_widget_destroy (options_window);
}

void
on_terrain_render_povray_activate      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  MainState *main_state;
  GtkWidget *terrain_window;
  GtkWidget *main_window;
  TTerrain  *terrain;
  gchar     *tmp, *base, *tga, *pov, *pov_safe, *out;
  gint       width, height;
  gchar     *command;
  GList     *cleanup_list;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  main_window =
    gtk_object_get_data (GTK_OBJECT (terrain_window), "data_parent");
  main_state =
    gtk_object_get_data (GTK_OBJECT (main_window), "data_state");

  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  base = filename_get_base (terrain->filename);
  tmp = g_strdup_printf ("/tmp/%s", base);
  g_free (base);

  pov = filename_new_extension (tmp, "pov");
  pov_safe = filename_povray_safe (pov);
  g_free (pov);
  t_terrain_save (terrain, FILE_POV, pov_safe);

  tga = filename_new_extension (pov_safe, "tga");
  t_terrain_save (terrain, FILE_TGA, tga);

  width = terrain->render_width;
  height = width * 3 / 4;

  base = filename_get_base (terrain->filename);
  out = filename_new_extension (pov_safe, "out.tga");
  g_free (tmp);

  command = g_strdup_printf ("%s +L%s +D +P +W%i +H%i +O%s +I%s",
                             main_state->povray_executable,
                             TERRAFORM_DATA_DIR "/include",
                             width, height, out, pov_safe);

  cleanup_list = NULL;
  cleanup_list = g_list_prepend (cleanup_list, out);
  cleanup_list = g_list_prepend (cleanup_list, pov_safe);
  cleanup_list = g_list_prepend (cleanup_list, tga);

  execute_command (command, _("There was an error executing POV-Ray"), cleanup,
                   cleanup_list);
  g_free (command);
}


void
on_terrain_native_activate             (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *terrain_window;
  GtkWidget *native_window;
  TTerrain  *terrain;
  TRgb      *rgb;
  SkyValues  sky;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  native_window = create_native_render_window ();
  rgb = t_rgb_new (lookup_widget (native_window, "image"),
                   terrain->render_width,
                   terrain->render_width * 3 / 4);
  gtk_object_set_data_full (GTK_OBJECT (native_window), "data_t_rgb", rgb,
                            (GtkDestroyNotify) t_rgb_free);

  gtk_widget_show (native_window);
  gtk_main_iteration ();

/*
  t_terrain_render (terrain, rgb,
                    EARTH_RADIUS, EARTH_CLOUD_HEIGHT, LENS_ANGLE,
                    0.0, 0.0, EARTH_RADIUS + 0.2);
*/

  sky_values_init (&sky);
  sky_render (rgb, &sky);
}


void
on_terrain_export_povray_activate      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *terrain_window;
  TTerrain  *terrain;
  gchar     *tga, *pov, *pov_safe;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  pov = filename_new_extension (terrain->filename, "pov");
  pov_safe = filename_povray_safe (pov);
  g_free (pov);
  t_terrain_save (terrain, FILE_POV, pov_safe);

  tga = filename_new_extension (pov_safe, "tga");
  g_free (pov_safe);
  t_terrain_save (terrain, FILE_TGA, tga);
  g_free (tga);
}


void
on_about_info_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *about_text;
  GtkWidget *about_frame;
  GtkWidget *copyright_label;
  gboolean   about_text_visible;

  about_text = lookup_widget (GTK_WIDGET (button), "about_text");
  about_frame = lookup_widget (GTK_WIDGET (button), "about_frame");
  copyright_label = GTK_BIN (button)->child;

  about_text_visible = GTK_WIDGET_VISIBLE (about_text);
  if (about_text_visible)
    {
      gtk_widget_hide (about_text);
      gtk_widget_show (about_frame);
      gtk_label_set_text (GTK_LABEL (copyright_label), _("Copyright Info"));
    }
  else
    {
      gtk_widget_hide (about_frame);
      gtk_widget_show (about_text);
      gtk_label_set_text (GTK_LABEL (copyright_label), _("Show Logo"));
    }
}


void
on_generic_cancel_clicked              (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_destroy (lookup_toplevel (GTK_WIDGET (button)));
}


static gchar *tool_group[] = {
  "move", "rectangular_select", "elliptical_select", NULL
};

static gchar *operator_group[] = {
  "replace", "add", "subtract", NULL
};

gboolean
on_terrain_button_release_event        (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
  GtkWidget *terrain_window;
  TTerrain  *terrain;
  TRgb      *rgb;

  terrain_window = lookup_toplevel (widget);
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");
  rgb = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_rgb");

  if (gtk_object_get_data (GTK_OBJECT (terrain_window), "data_anchor") != NULL)
    {
      if (terrain->view == T_VIEW_2D_PLANE && 
          !get_boolean (terrain_window, "move"))
        {
          TSelectionType type;
          TComposeOp     op;
          gfloat         x1, y1, x2, y2;

          type = t_rgb_get_selection (rgb, &x1, &y1, &x2, &y2);
          op = (TComposeOp)
            (get_radio_menu (terrain_window, operator_group) + 1);

          t_terrain_select (terrain, x1, y1, x2, y2, type, op);

          t_rgb_set_selection (rgb, T_SELECTION_NONE, 0, 0, 0, 0);
          terrain_window_queue_paint (terrain_window);
        }

      gtk_object_remove_data (GTK_OBJECT (terrain_window), "data_anchor");
      gtk_grab_remove (widget);
    }

  if (gtk_object_get_data (GTK_OBJECT (terrain_window), "data_object") != NULL)
    {
      gint   *item;
      gfloat  x, y, angle;

      item = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_object");
      t_rgb_get_object (rgb, *item, &x, &y, &angle);
      t_terrain_move_object (terrain, *item, x, y);
      t_terrain_rotate_object (terrain, *item, angle);

      gtk_object_remove_data (GTK_OBJECT (terrain_window), "data_object");
    }

  return FALSE;
}

gboolean
on_terrain_motion_notify_event         (GtkWidget       *widget,
                                        GdkEventMotion  *event,
                                        gpointer         user_data)
{
  GtkWidget *terrain_window;
  TTerrain  *terrain;
  TRgb      *rgb;
  gint      *anchor;
  gint       x, y;

  /* Get x and y coordinates on hint. */
  if (event->is_hint)
    {
      gint xint, yint;
      GdkModifierType state;

      gdk_window_get_pointer (event->window, &xint, &yint, &state);
      x = xint;
      y = yint;
    }
  else
    x = event->x, y = event->y;

  terrain_window = lookup_toplevel (widget);
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");
  rgb = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_rgb");
  anchor = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_anchor");

  if (anchor != NULL && terrain->view == T_VIEW_2D_PLANE)
    {
      gint  tool;
      gint *item;

      tool = get_radio_menu (terrain_window, tool_group);

      switch (tool)
        {
        case 0: /* MOVE */
          item =
            gtk_object_get_data (GTK_OBJECT (terrain_window), "data_object");

          if (item != NULL)
            {
              TTerrainObject *object;
              GArray         *array;
              gfloat          obj_x, obj_y;
              gfloat          delta_x, delta_y;

              delta_x = ((gfloat) (x - anchor[0])) / (rgb->width - 1);
              delta_y = ((gfloat) (y - anchor[1])) / (rgb->height - 1);

              array = terrain->objects;
              object = &g_array_index (array, TTerrainObject, *item);
              obj_x = CLIP (object->x + delta_x, 0.0, 1.0);
              obj_y = CLIP (object->y + delta_y, 0.0, 1.0);

              t_rgb_modify_object (rgb, *item, obj_x, obj_y, object->angle);
            }

          break;

        case 1: /* RECTANGLE */
          t_rgb_set_selection (rgb, T_SELECTION_RECTANGLE,
                               anchor[0], anchor[1], x, y);
          break;

        case 2: /* ELLIPSE */
          t_rgb_set_selection (rgb, T_SELECTION_ELLIPSE,
                               anchor[0], anchor[1], x, y);
          break;
        }
    }
  else if (anchor != NULL && terrain->view == T_VIEW_3D_WIRE)
    {
      terrain->angle += (x - anchor[0]) * 0.005;
      while (terrain->angle < 0.0) terrain->angle += 2 * M_PI;
      while (terrain->angle >= 2 * M_PI) terrain->angle -= 2 * M_PI;

      terrain->elevation += (y - anchor[1]) * 0.004;
      if (terrain->elevation < -2.0) terrain->elevation = -2.0;
      if (terrain->elevation > 2.0) terrain->elevation = 2.0;

      anchor = g_new (guint, 2);
      anchor[0] = event->x;
      anchor[1] = event->y;
      gtk_object_set_data_full (GTK_OBJECT (terrain_window),
                                "data_anchor", anchor, g_free);

      t_terrain_paint (terrain, rgb);
      t_rgb_hard_sync (rgb);
    }

  return TRUE;
}


void
on_option_download_povray_clicked      (GtkButton       *button,
                                        gpointer         user_data)
{

}


GtkWidget*
create_options_download_povray_url (gchar *widget_name, gchar *string1, gchar *string2,
                gint int1, gint int2)
{
  return gnome_href_new ("http://www.povray.org/", _("Download POV-Ray..."));
}


static gboolean
render_static (gpointer data)
{
  TRgb *rgb;

  rgb = gtk_object_get_data (GTK_OBJECT (data), "data_t_rgb");
  if (rgb != NULL)
    {
      t_rgb_render_static (rgb);
      t_rgb_hard_sync (rgb);
    }

  rgb = gtk_object_get_data (GTK_OBJECT (data), "data_function_t_rgb");
  if (rgb != NULL)
    {
      t_rgb_render_static (rgb);
      t_rgb_hard_sync (rgb);
    }

  return TRUE;
}

void
on_generic_use_preview_toggled         (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  GtkWidget *filter_window;
  TRgb *rgb;

  filter_window = lookup_toplevel (GTK_WIDGET (togglebutton));
  rgb = gtk_object_get_data (GTK_OBJECT (filter_window), "data_t_rgb");
  if (get_boolean (GTK_WIDGET (filter_window), "use_preview"))
    {
      filter_window_update_preview (filter_window);
      widget_remove_timer (filter_window);
    }
  else
    {
      t_rgb_render_static (rgb);
      widget_set_timer (filter_window, 100, (GtkFunction) render_static);
    }

  t_rgb_soft_sync (rgb);
}


void
on_render_apply_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *options_window;
  GtkWidget *terrain_window;
  TTerrain  *terrain;

  options_window = lookup_toplevel (GTK_WIDGET (button));
  terrain_window = gtk_object_get_data (GTK_OBJECT (options_window),
                                        "data_parent");
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  t_terrain_pack_options (terrain, options_window);

  terrain_window_refresh (terrain_window, FALSE);
}


void
on_save_as_ok_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget    *save_as;
  GtkWidget    *terrain_window;
  GtkWidget    *main_window;
  TTerrain     *terrain;
  gchar        *filename;
  gint          status;
  TFileType     type;
  FileTypeInfo *info;

  save_as = lookup_toplevel (GTK_WIDGET (button));
  terrain_window = gtk_object_get_data (GTK_OBJECT (save_as), "data_parent");
  main_window = gtk_object_get_data (GTK_OBJECT (terrain_window),
                                     "data_parent");
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");
  filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (save_as));
  info = gtk_object_get_data (GTK_OBJECT (save_as), "data_file_type_info");
  type = info[get_option (save_as, "type_menu")].type;
  status = t_terrain_save (terrain, type, filename);
  terrain_window_set_title (terrain_window, terrain_window);
  main_window_update_terrain_window (main_window, terrain_window);

  gtk_widget_destroy (save_as);

  if (status != 0)
    {
      GtkWidget *error_window;

      error_window = create_error_window ();
      set_text (error_window, "error_label", _("There was an error while saving the file."));
      gtk_widget_show (error_window);
    }
}


void
on_use_preview_toggled                 (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{

}


gboolean
on_render_custom_button_press_event    (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
  TTerrain  *terrain;
  GtkWidget *render_options;
  TRgb      *rgb;
  gfloat     elevation_offset;
  gfloat     y;

  render_options = lookup_toplevel (widget);
  rgb = gtk_object_get_data (GTK_OBJECT (render_options), "data_t_rgb");
  terrain = gtk_object_get_data (GTK_OBJECT (render_options), "data_t_terrain");

  elevation_offset = get_float (widget, "render_elevation");
  y = terrain->heightfield[((gint) event->y) *
      terrain->width + ((gint) event->x)];

  if (event->button == 1)
    {
      /* Left button. */
      if (get_boolean (widget, "render_observe_sealevel"))
        y = MAX (y, terrain->sealevel);

      set_float (widget, "render_camera_x", ((gfloat) event->x) / rgb->width);
      set_float (widget, "render_camera_y", y + elevation_offset);
      set_float (widget, "render_camera_z", ((gfloat) event->y) / rgb->height);
    }
  else if (event->button == 2)
    {
      /* Middle button. */
      if (get_boolean (widget, "render_observe_sealevel"))
        y = MAX (y, terrain->sealevel);

      set_float (widget, "render_lookat_x", ((gfloat) event->x) / rgb->width);
      set_float (widget, "render_lookat_y", y + elevation_offset);
      set_float (widget, "render_lookat_z", ((gfloat) event->y) / rgb->height);
    }
  else if (event->button == 3)
    {
      /* Right button. */
      on_generic_preview_button_press_event (widget, event, user_data);
    }

  return FALSE;
}


static void
preview_popup_destroy (gpointer user_data)
{
  gtk_object_remove_data (GTK_OBJECT (user_data), "data_popup");
}


gboolean
on_generic_preview_button_press_event  (GtkWidget       *widget,
                                        GdkEventButton  *event,
                                        gpointer         user_data)
{
  GtkWidget *window;
  GtkWidget *center_x_widget;
  TRgb      *rgb;
  gfloat     center_x, center_y;

  window = lookup_toplevel (widget);

  if (event->button == 1)
    {
      center_x_widget = gtk_object_get_data (GTK_OBJECT (window), "center_x");
      if (center_x_widget == NULL)
        return FALSE;

      if (!GTK_WIDGET_IS_SENSITIVE (center_x_widget))
        return FALSE;

      rgb = gtk_object_get_data (GTK_OBJECT (window), "data_t_rgb");
      center_x = ((gfloat) event->x) / rgb->width;
      center_y = ((gfloat) event->y) / rgb->height;

      gtk_object_set_data (GTK_OBJECT (window), "dont_update_preview", window);
      set_float (window, "center_x", center_x);
      gtk_object_remove_data (GTK_OBJECT (window), "dont_update_preview");
      set_float (window, "center_y", center_y);
    }
  else if (event->button == 3)
    {
      GtkWidget *preview_popup;
      GtkWidget *use_preview;

      use_preview = gtk_object_get_data (GTK_OBJECT (window), "use_preview");
      if (use_preview != NULL && !get_boolean (window, "use_preview"))
        return FALSE;

      preview_popup = gtk_object_get_data (GTK_OBJECT (window), "data_popup");
      if (preview_popup == NULL)
        {
          TTerrain *terrain;

          preview_popup = create_preview_popup ();

          gtk_object_set_data_full (
            GTK_OBJECT (preview_popup), "data_parent", window,
            (GtkDestroyNotify) preview_popup_destroy);

          gtk_object_set_data (
            GTK_OBJECT (window), "data_popup", preview_popup);

          terrain = gtk_object_get_data (GTK_OBJECT (window), "data_t_terrain");
          t_terrain_unpack_preview_menu (terrain, preview_popup);

          gtk_menu_popup (GTK_MENU (preview_popup), NULL, NULL, NULL, NULL, 3,
                          event->time);
        }
    }

  return FALSE;
}


gboolean
on_generic_preview_motion_notify_event (GtkWidget       *widget,
                                        GdkEventMotion  *event,
                                        gpointer         user_data)
{
  GtkWidget *window;
  GtkWidget *center_x_widget;
  TRgb      *rgb;
  gfloat     center_x, center_y;
  gint       x, y;

  window = lookup_toplevel (widget);

  center_x_widget = gtk_object_get_data (GTK_OBJECT (window), "center_x");
  if (center_x_widget == NULL)
    return FALSE;

  if (!GTK_WIDGET_IS_SENSITIVE (center_x_widget))
    return FALSE;

  rgb = gtk_object_get_data (GTK_OBJECT (window), "data_t_rgb");
  if (event->is_hint)
    {
      gint xint, yint;
      GdkModifierType state;

      gdk_window_get_pointer (event->window, &xint, &yint, &state);
      x = xint;
      y = yint;
    }
  else
    {
      x = event->x;
      y = event->y;
    }

  center_x = ((gfloat) x) / rgb->width;
  center_y = ((gfloat) y) / rgb->height;

  gtk_object_set_data (GTK_OBJECT (window), "dont_update_preview", window);
  set_float (window, "center_x", center_x);
  gtk_object_remove_data (GTK_OBJECT (window), "dont_update_preview");
  set_float (window, "center_y", center_y);

  return FALSE;
}


void
on_generic_toggled                     (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  filter_window_update_preview (lookup_toplevel (GTK_WIDGET (togglebutton)));
}


void
on_generic_ok_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
  filter_window_perform (lookup_toplevel (GTK_WIDGET (button)));
}


void
on_tile_preview_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{

}

void
on_preview_popup_selection_done        (GtkMenuShell    *menushell,
                                        gpointer         user_data)
{
  GtkWidget *filter_window;
  TTerrain  *terrain;

  filter_window = gtk_object_get_data (GTK_OBJECT (menushell), "data_parent");
  terrain = gtk_object_get_data (GTK_OBJECT (filter_window), "data_t_terrain");
  t_terrain_pack_preview_menu (terrain, GTK_WIDGET (menushell));
  gtk_object_unref (GTK_OBJECT (menushell));
}


void
on_undo_ok_clicked                     (GtkButton       *button,
                                        gpointer         user_data)
{
  TTerrain  *terrain;
  TRgb      *rgb;
  GtkWidget *undo_window;
  GtkWidget *terrain_window;
  GtkWidget *main_window;
  gfloat     scale;

  undo_window = lookup_toplevel (GTK_WIDGET (button));
  terrain_window = gtk_object_get_data (GTK_OBJECT (undo_window),
                                        "data_parent");
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window),
                                 "data_t_terrain");
  rgb = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_rgb");

  main_window = gtk_object_get_data (GTK_OBJECT (terrain_window),
                                     "data_parent");

  undo_apply (undo_window, terrain);

  scale = ((gfloat) TERRAIN_MIN_VIEW) / MAX (terrain->width, terrain->height);
  t_rgb_set_size (rgb, scale * terrain->width, scale * terrain->height);

  terrain->modified = TRUE;
  terrain_window_queue_paint (terrain_window);
  terrain_window_refresh_contours (terrain_window);

  gtk_widget_destroy (undo_window);
}


void
on_tree_selection_changed              (GtkTree         *tree,
                                        gpointer         user_data)
{
  GtkWidget *undo_window;
  GtkTree   *root_tree;
  GList     *selection;
  gboolean   sensitive;

  undo_window = lookup_toplevel (GTK_WIDGET (tree));
  root_tree = GTK_TREE (lookup_widget (undo_window, "tree"));

  selection = GTK_TREE_SELECTION (root_tree);
  sensitive =
    selection != NULL &&
    gtk_object_get_data (GTK_OBJECT (selection->data), "data_state") != NULL;

  gtk_widget_set_sensitive (lookup_widget (GTK_WIDGET (undo_window), "undo_ok"),
                            sensitive);

  if (sensitive)
    gtk_object_remove_data (GTK_OBJECT (undo_window), "show_black");
  else
    gtk_object_set_data (GTK_OBJECT (undo_window),
                         "show_black", &undo_window);

  filter_window_update_preview (undo_window);
}


void
on_execution_error_details_click       (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *error_log;
  gboolean   visible;

  error_log = lookup_widget (GTK_WIDGET (button), "error_log");
  visible = GTK_WIDGET_VISIBLE (error_log);
  if (visible)
    gtk_widget_hide (error_log);
  else
    gtk_widget_show (error_log);

  if (visible)
    set_text (GTK_WIDGET (button), "details", _("Show Details"));
  else
    set_text (GTK_WIDGET (button), "details", _("Hide Details"));
}


void
on_terrain_print_activate              (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GnomePrinter      *printer;
  GnomePrintContext *context;
  GtkWidget         *terrain_window;
  TTerrain          *terrain;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  printer = gnome_printer_dialog_new_modal ();

  if (printer != NULL)
    {
      context = gnome_print_context_new (printer);
      gnome_print_beginpage (context, "Terrain");
      t_terrain_print_contour_map (terrain, context);
      gnome_print_showpage (context);
      gnome_print_context_close (context);
      gtk_object_unref (GTK_OBJECT (context));
      gtk_object_unref (GTK_OBJECT (printer));
    }
}

void
on_erode_save_toggled                  (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  GtkWidget *widget;
  gboolean enabled;

  widget = GTK_WIDGET (togglebutton);
  enabled = get_boolean (widget, "erode_save");
  gtk_widget_set_sensitive (lookup_widget (widget, "erode_file"), enabled);
  gtk_widget_set_sensitive (lookup_widget (widget, "erode_frame_count_label"),
                            enabled);
  gtk_widget_set_sensitive (lookup_widget (widget, "erode_frame_count"),
                            enabled);
}


void
on_merge_ok_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *merge_window;

  merge_window = lookup_toplevel (GTK_WIDGET (button));
  merge_apply (merge_window, FALSE);
  gtk_widget_destroy (merge_window);
}


void
on_merge_cancel_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
  gtk_widget_destroy (lookup_toplevel (GTK_WIDGET (button)));
}


void
on_merge_list_select_row               (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  GtkWidget *widget;
  GtkWidget *merge_window;
  gint      *row_number;

  row_number = g_new (gint, 1);
  *row_number = row;
  merge_window = lookup_toplevel (GTK_WIDGET (clist));
  gtk_object_set_data_full (GTK_OBJECT (merge_window), "data_row",
                            row_number, (GtkDestroyNotify) g_free);

  widget = lookup_widget (GTK_WIDGET (clist), "top_select");
  gtk_widget_set_sensitive (widget, TRUE);
  widget = lookup_widget (GTK_WIDGET (clist), "bottom_select");
  gtk_widget_set_sensitive (widget, TRUE);
}


void
on_merge_list_unselect_row             (GtkCList        *clist,
                                        gint             row,
                                        gint             column,
                                        GdkEvent        *event,
                                        gpointer         user_data)
{
  GtkWidget *widget;

  widget = lookup_widget (GTK_WIDGET (clist), "top_select");
  gtk_widget_set_sensitive (widget, FALSE);
  widget = lookup_widget (GTK_WIDGET (clist), "bottom_select");
  gtk_widget_set_sensitive (widget, FALSE);
}


void
on_top_select_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
  TRgb      *rgb;
  GtkWidget *main_window;
  GtkWidget *terrain_window;
  GtkWidget *merge_window;
  GtkWidget *frame;
  TTerrain  *terrain;
  gint       row;
  gchar     *text;

  merge_window = lookup_toplevel (GTK_WIDGET (button));
  row = *((gint*) gtk_object_get_data (GTK_OBJECT (merge_window), "data_row"));

  main_window = lookup_widget (merge_window, "data_parent");
  terrain_window = main_window_get_terrain_window (main_window, row);
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");
  t_terrain_ref (terrain);
  gtk_object_set_data_full (GTK_OBJECT (merge_window),
                            "data_top_t_terrain", terrain,
                            (GtkDestroyNotify) t_terrain_unref);
  terrain = t_terrain_clone_preview (terrain);
  gtk_object_set_data_full (GTK_OBJECT (merge_window),
                            "data_top_thumbnail", terrain,
                            (GtkDestroyNotify) t_terrain_unref);
  frame = lookup_widget (merge_window, "top_frame");
  text = terrain_window_get_title (terrain_window);
  gtk_frame_set_label (GTK_FRAME (frame), text);
  g_free (text);

  rgb = gtk_object_get_data (GTK_OBJECT (merge_window), "data_top_t_rgb");
  t_terrain_paint_preview (terrain, rgb);
  t_rgb_soft_sync (rgb);

  merge_apply (merge_window, TRUE);
}


void
on_bottom_select_clicked               (GtkButton       *button,
                                        gpointer         user_data)
{
  TRgb      *rgb;
  GtkWidget *main_window;
  GtkWidget *terrain_window;
  GtkWidget *merge_window;
  GtkWidget *frame;
  TTerrain  *terrain;
  gint       row;
  gchar     *text;

  merge_window = lookup_toplevel (GTK_WIDGET (button));
  row = *((gint*) gtk_object_get_data (GTK_OBJECT (merge_window), "data_row"));

  main_window = lookup_widget (merge_window, "data_parent");
  terrain_window = main_window_get_terrain_window (main_window, row);
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");
  t_terrain_ref (terrain);
  gtk_object_set_data_full (GTK_OBJECT (merge_window),
                            "data_bottom_t_terrain", terrain,
                            (GtkDestroyNotify) t_terrain_unref);
  terrain = t_terrain_clone_preview (terrain);
  gtk_object_set_data_full (GTK_OBJECT (merge_window),
                            "data_bottom_thumbnail", terrain,
                            (GtkDestroyNotify) t_terrain_unref);
  frame = lookup_widget (merge_window, "bottom_frame");
  text = terrain_window_get_title (terrain_window);
  gtk_frame_set_label (GTK_FRAME (frame), text);
  g_free (text);

  rgb = gtk_object_get_data (GTK_OBJECT (merge_window), "data_bottom_t_rgb");
  t_terrain_paint_preview (terrain, rgb);
  t_rgb_soft_sync (rgb);

  merge_apply (merge_window, TRUE);
}


void
on_native_save_clicked                 (GtkButton       *button,
                                        gpointer         user_data)
{

}


void
on_scale_ok_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
  TTerrain  *terrain;
  GtkWidget *terrain_window;
  GtkWidget *scale_window;
  gboolean   double_size;

  scale_window = lookup_toplevel (GTK_WIDGET (button));
  terrain_window = lookup_widget (scale_window, "data_parent");
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  double_size = get_boolean (GTK_WIDGET (button), "double_size");

  if (double_size)
    {
      gfloat local_scaling;
      gfloat global_scaling;

      local_scaling = get_float (scale_window, "local_scaling");
      global_scaling = get_float (scale_window, "global_scaling");
      t_terrain_double (terrain, local_scaling, global_scaling);
    }
  else
    t_terrain_half (terrain);

  terrain_window_save_state (terrain_window, _("Scale"));
  terrain_window_refresh (terrain_window, FALSE);

  gtk_widget_destroy (scale_window);
}

void
on_scale_toggled                       (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  gboolean double_size;

  double_size = get_boolean (GTK_WIDGET (togglebutton), "double_size");
  gtk_widget_set_sensitive (
    lookup_widget (GTK_WIDGET (togglebutton),
                   "local_label"), double_size);
  gtk_widget_set_sensitive (
    lookup_widget (GTK_WIDGET (togglebutton),
                   "local_scaling"), double_size);
  gtk_widget_set_sensitive (
    lookup_widget (GTK_WIDGET (togglebutton),
                   "global_label"), double_size);
  gtk_widget_set_sensitive (
    lookup_widget (GTK_WIDGET (togglebutton),
                   "global_scaling"), double_size);
}


void
on_terrain_print_preview_activate      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GnomePrintContext *context;
  GtkWidget         *preview_window;
  GtkWidget         *terrain_window;
  GtkWidget         *canvas;
  TTerrain          *terrain;

  preview_window = create_print_preview_window ();

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  canvas = lookup_widget (preview_window, "preview_canvas");

  context = gnome_print_preview_new (GNOME_CANVAS (canvas), "US-Letter");
  
  gnome_print_beginpage (context, "Terrain");
  t_terrain_print_contour_map (terrain, context);
  gnome_print_showpage (context);
  gnome_print_context_close (context);

  gtk_widget_show (preview_window);
}

void
on_select_invert_activate              (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *terrain_window;
  TTerrain  *terrain;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  t_terrain_select_invert (terrain);
  if (terrain->view == T_VIEW_2D_PLANE)
    terrain_window_queue_paint (terrain_window);
}


void
on_select_all_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *terrain_window;
  TTerrain  *terrain;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  t_terrain_select_all (terrain);
  if (terrain->view == T_VIEW_2D_PLANE)
    terrain_window_queue_paint (terrain_window);
}


void
on_select_none_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *terrain_window;
  TTerrain  *terrain;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  if (terrain->selection != NULL)
    {
      t_terrain_select_none (terrain);
      if (terrain->view == T_VIEW_2D_PLANE)
        terrain_window_queue_paint (terrain_window);
    }
}


static void
feather_apply                          (GtkWidget       *window,
                                        TTerrain        *terrain)
{
  gint radius;

  radius = get_int (window, "radius");
  t_terrain_select_feather (terrain, radius);
}


void
on_select_feather_activate             (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = { NULL };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_feather_window,
                     &feather_apply,
                     hscales,
                     FALSE);
}


static void
select_by_height_apply                 (GtkWidget       *window,
                                        TTerrain        *terrain)
{
  gfloat the_floor, the_ceiling;

  the_floor = get_float (window, "floor");
  the_ceiling = get_float (window, "ceiling");

  t_terrain_select_by_height (terrain, the_floor, the_ceiling);
}


void
on_by_height_activate                  (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *hscales[] = { "floor", "ceiling", NULL };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_select_by_height_window,
                     &select_by_height_apply,
                     hscales,
                     FALSE);
}


void
on_level_connector_count_changed       (GtkEditable     *editable,
                                        gpointer         user_data)
{
  filter_window_update_preview (lookup_toplevel (GTK_WIDGET (editable)));
}


void
place_objects_apply                    (GtkWidget       *window,
                                        TTerrain        *terrain)
{
  GList    *object_names;
  gchar    *name;
  gfloat    elevation_min, elevation_max;
  gfloat    density, variance;
  gfloat    scale_x, scale_y, scale_z;
  gfloat    variance_x, variance_y, variance_z;
  gboolean  vary_direction;
  gboolean  proportional;
  gint      seed;

  object_names = gtk_object_get_data (GTK_OBJECT (window), "data_objects");
  name = g_list_nth_data (object_names, get_option (window, "object"));
  elevation_min = get_float (window, "elevation_min");
  elevation_max = get_float (window, "elevation_max");
  density = get_float (window, "density");
  variance = get_float (window, "variance");
  scale_x = get_float (window, "scale_x");
  scale_y = get_float (window, "scale_y");
  scale_z = get_float (window, "scale_z");
  variance_x = get_float (window, "variance_x");
  variance_y = get_float (window, "variance_y");
  variance_z = get_float (window, "variance_z");
  vary_direction = get_boolean (window, "vary_direction");
  proportional = get_boolean (window, "proportional");

  if (get_boolean (window, "new_seed"))
    seed = -1;
  else
    seed = get_int (window, "seed");

  t_terrain_place_objects (terrain, name,
    elevation_min, elevation_max, density, variance,
    scale_x, scale_y, scale_z, variance_x, variance_y, variance_z,
    vary_direction,
    proportional,
    seed);
}


static GtkWidget *
create_place_window_advanced           (TTerrain        *terrain)
{
  gchar     *text;
  gfloat     pov_x, pov_y, pov_z;
  GtkWidget *place_window;
  GtkWidget *object_menu;
  GList     *list;

  place_window = create_place_objects_window ();

  list = object_list_new ();
  object_menu = gtk_menu_new ();
  if (list == NULL)
    list = g_list_append (list, g_strdup (_("[No Objects Installed]")));

  gtk_object_set_data_full (GTK_OBJECT (place_window),
                            "data_objects", list,
                            (GtkDestroyNotify) file_list_free);

  t_terrain_get_povray_size (terrain, &pov_x, &pov_y, &pov_z);
  text = g_strdup_printf (_("The terrain size is (%.2f,%.2f,%.2f)"), pov_x, pov_y, pov_z);
  set_text (place_window, "terrain_info", text);
  g_free (text);

  set_float (place_window, "elevation_min", terrain->sealevel);
  set_float (place_window, "elevation_max", (1.0 + terrain->sealevel) * 0.5);

  for (; list != NULL; list = list->next)
    {
      GtkWidget *object_menuitem;
      guchar    *friendly_name;

      friendly_name = filename_get_friendly ((gchar*) list->data);
      object_menuitem = gtk_menu_item_new_with_label (friendly_name);
      g_free (friendly_name);

      gtk_widget_show (object_menuitem);
      gtk_menu_append (GTK_MENU (object_menu), object_menuitem);
    }

  gtk_option_menu_set_menu (
    GTK_OPTION_MENU (lookup_widget (place_window, "object")),
    object_menu);

  return place_window;
}


void
on_place_activate                      (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  /*
   * Even though the place objects window contains hscales, it does not have
   * a preview.  Therefore, we don't bother adding update signal
   * callbacks to the hscales.
   */

  gchar *hscales[] = { NULL };

  filter_window_new (GTK_WIDGET (menuitem),
                     (CreateWindowFunc) &create_place_window_advanced,
                     &place_objects_apply,
                     hscales,
                     FALSE);
}


void
on_remove_all_activate                 (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  GtkWidget *terrain_window;
  TTerrain  *terrain;

  terrain_window = lookup_toplevel (GTK_WIDGET (menuitem));
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  t_terrain_remove_objects (terrain);
  terrain_window_refresh_objects (terrain_window);
}


void
on_generic_radio_toggled               (GtkToggleButton *togglebutton,
                                        gpointer         user_data)
{
  if (gtk_toggle_button_get_active (togglebutton))
    filter_window_update_preview (lookup_toplevel (GTK_WIDGET (togglebutton)));
}


void
on_place_ok_clicked                    (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *filter_window;
  GtkWidget *terrain_window;

  filter_window = lookup_toplevel (GTK_WIDGET (button));
  terrain_window = lookup_widget (filter_window, "data_parent");
  filter_window_perform (filter_window);
  terrain_window_refresh_objects (terrain_window);
}


static void
goto_page (gchar *page)
{
  gchar *path1, *path2;

  path1 = gnome_help_file_path ("terraform", page);
  path2 = g_strdup_printf ("ghelp:%s", path1);
  g_free (path1);
  gnome_help_goto (NULL, path2);
  g_free (path2);
}


void
on_tutorial_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  goto_page ("tutorial.html");
}


void
on_users_guide_activate                (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  goto_page ("index.html");
}

void
on_sub_generate_clicked                (GtkButton       *button,
                                        gpointer         user_data)
{
  TTerrain  *terrain;
  GtkWidget *subdivision_window, *main_window;
  gint       size;
  gfloat     scale;
  gint       seed;
  gint       method;

  subdivision_window = lookup_toplevel (GTK_WIDGET (button));
  main_window = lookup_widget (GTK_WIDGET (subdivision_window), "data_parent");

  method = get_option (subdivision_window, "sub_method");
  size = get_int (subdivision_window, "sub_size");
  scale = get_float (subdivision_window, "sub_scale_factor"); 
  if (get_boolean (subdivision_window, "new_seed"))
    seed = -1;
  else
    seed = get_int (subdivision_window, "seed");

  terrain = t_terrain_generate_subdiv (method, size, scale, seed);

  if (terrain != NULL)
    {
      GtkWidget *terrain_window;

      terrain_window = gtk_object_get_data (GTK_OBJECT (subdivision_window),
                                            "data_terrain_window");

      if (terrain_window == NULL)
        {
          terrain_window = terrain_window_new (main_window, terrain);
          set_text (subdivision_window, "generate", _("Regenerate"));
          set_text (subdivision_window, "cancel", _("Finish"));
          gtk_object_set_data (GTK_OBJECT (subdivision_window),
                               "data_terrain_window", terrain_window);
        }
      else
        terrain_window_set_terrain (terrain_window, terrain);

      terrain_window_save_state (terrain_window, _("Subdivision Synthesis"));
      t_terrain_unref (terrain);
    }
}


void
on_sub_cancel_clicked                  (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *subdivision_window;

  subdivision_window = lookup_toplevel (GTK_WIDGET (button));

  gtk_widget_destroy (subdivision_window);
}


void
on_ss_generate_clicked                 (GtkButton       *button,
                                        gpointer         user_data)
{
  TTerrain  *terrain;
  GtkWidget *spectral_window, *main_window;
  gint       size;
  gfloat     dimensions;
  gboolean   invert;
  gint       seed;

  spectral_window = lookup_toplevel (GTK_WIDGET (button));
  main_window = lookup_widget (GTK_WIDGET (spectral_window), "data_parent");

  size = get_int (spectral_window, "ss_size");
  dimensions = get_float (spectral_window, "ss_fractal_dimensions"); 
  invert = get_boolean (spectral_window, "ss_invert_trig");
  if (get_boolean (spectral_window, "new_seed"))
    seed = -1;
  else
    seed = get_int (spectral_window, "seed");

  terrain = t_terrain_generate_spectral (size, 3.0 - dimensions, seed, invert);

  if (terrain != NULL)
    {
      GtkWidget *terrain_window;

      terrain_window = gtk_object_get_data (GTK_OBJECT (spectral_window),
                                            "data_terrain_window");

      if (terrain_window == NULL)
        {
          terrain_window = terrain_window_new (main_window, terrain);
          set_text (spectral_window, "generate", _("Regenerate"));
          set_text (spectral_window, "cancel", _("Finish"));
          gtk_object_set_data (GTK_OBJECT (spectral_window),
                               "data_terrain_window", terrain_window);
        }
      else
        terrain_window_set_terrain (terrain_window, terrain);

      terrain_window_save_state (terrain_window, _("Spectral Synthesis"));
      t_terrain_unref (terrain);
    }
}


void
on_ss_cancel_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *spectral_window;

  spectral_window = lookup_toplevel (GTK_WIDGET (button));

  gtk_widget_destroy (spectral_window);
}


void
on_pn_generate_clicked                 (GtkButton       *button,
                                        gpointer         user_data)
{
  TTerrain  *terrain;
  GtkWidget *widget;
  GtkWidget *pn_window;
  GtkWidget *main_window;
  gboolean   enabled[4];
  gfloat     frequency[4];
  gfloat     amplitude[4];
  gint       iterations[4];
  gint       filter[4];
  gint       size;
  gint       seed;
  gint       i;

  widget = GTK_WIDGET (button);
  pn_window = lookup_toplevel (GTK_WIDGET (button));
  main_window = lookup_widget (GTK_WIDGET (pn_window), "data_parent");

  for (i = 0; i < 4; i++)
    {
      enabled[i] = get_boolean (widget, perlin_rows[i][0]);
      frequency[i] = get_float (widget, perlin_rows[i][1]);
      amplitude[i] = get_float (widget, perlin_rows[i][2]);
      iterations[i] = get_int (widget, perlin_rows[i][3]);
      filter[i] = get_option (widget, perlin_rows[i][4]);
    }

  size = get_int (widget, "pn_size");

  if (get_boolean (widget, "new_seed"))
    seed = -1;
  else
    seed = get_int (widget, "seed");

  terrain =
    t_terrain_generate_perlin (size, size, seed, enabled,
                               frequency, amplitude, iterations, filter);

  if (terrain != NULL)
    {
      GtkWidget *terrain_window;

      terrain_window =
        gtk_object_get_data (GTK_OBJECT (pn_window), "data_terrain_window");

      if (terrain_window == NULL)
        {
          terrain_window = terrain_window_new (main_window, terrain);
          set_text (pn_window, "generate", _("Regenerate"));
          set_text (pn_window, "cancel", _("Finish"));
          gtk_object_set_data (GTK_OBJECT (pn_window),
                               "data_terrain_window", terrain_window);
        }
      else
        terrain_window_set_terrain (terrain_window, terrain);

      terrain_window_save_state (terrain_window, _("Perlin Noise"));
      t_terrain_unref (terrain);
    }
}



void
on_pn_cancel_clicked                   (GtkButton       *button,
                                        gpointer         user_data)
{
  GtkWidget *perlin_window;

  perlin_window = lookup_toplevel (GTK_WIDGET (button));

  gtk_widget_destroy (perlin_window);
}
