/* Terraform - (C) 1997-2000 Robert Gasch (r.gasch@chello.nl)
 *  - http://terraform.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "contour.h"
#include "filterwindow.h"
#include "support.h"
#include "support2.h"
#include "terrainwindow.h"

/* Helper functions. */

/*
 * filter_window_update_preview: Update a filter's preview window
 */

void
filter_window_update_preview (GtkWidget *filter_window)
{
  TRgb        *rgb;
  TTerrain    *terrain;
  TTerrain    *reference;
  GtkWidget   *widget;

  if (gtk_object_get_data (GTK_OBJECT (filter_window), "use_preview") != NULL)
    if (!get_boolean (filter_window, "use_preview"))
      return;

  if (gtk_object_get_data (GTK_OBJECT (filter_window), "preview") == NULL)
    return;

  rgb = gtk_object_get_data (GTK_OBJECT (filter_window), "data_t_rgb");
  terrain = gtk_object_get_data (GTK_OBJECT (filter_window), "data_t_terrain");
  reference = gtk_object_get_data (GTK_OBJECT (filter_window),
                                   "data_t_terrain_reference");

  if (reference != NULL)
    {
      t_terrain_set_size (terrain, reference->width, reference->height);
      t_terrain_copy_heightfield_and_selection (reference, terrain);
    }


  if (gtk_object_get_data (GTK_OBJECT (filter_window), "show_black"))
    {
      t_rgb_clear (rgb);
    }
  else
    {
      FilterApply  filter_apply;

      filter_apply = gtk_object_get_data (GTK_OBJECT (filter_window),
                                      "data_filter_apply_func");
      if (filter_apply != NULL)
        filter_apply (filter_window, terrain);
      t_terrain_contour_lines (terrain, terrain->contour_levels, 5);

      t_rgb_set_size (rgb, terrain->width, terrain->height);
      gtk_widget_queue_resize (filter_window);
      t_terrain_paint (terrain, rgb);
    }

  widget = gtk_object_get_data (GTK_OBJECT (filter_window), "center_x");
  if (widget != NULL && GTK_WIDGET_IS_SENSITIVE (widget))
    {
      gfloat center_x, center_y;
      gint x, y;

      center_x = get_float (filter_window, "center_x");
      center_y = get_float (filter_window, "center_y");

      x = center_x * rgb->width;
      y = center_y * rgb->height;
      t_rgb_set_fore (rgb, 0, 0, 0);
      t_rgb_draw_line (rgb, 0, y, rgb->width - 1, y);
      t_rgb_draw_line (rgb, x, 0, x, rgb->height - 1);
    }

  t_rgb_hard_sync (rgb);
}


/*
 * filter_window_perform: Perform filter on the terrain and close
 * the filter window
 */

void
filter_window_perform (GtkWidget *filter_window)
{
  FilterApply  filter_apply;
  TTerrain    *terrain;
  TRgb        *rgb;
  GtkWidget   *terrain_window;
  GtkWidget   *main_window;
  gchar       *filter_name;
  gfloat       scale;

  terrain_window = gtk_object_get_data (GTK_OBJECT (filter_window),
                                        "data_parent");
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");
  rgb = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_rgb");
  main_window = gtk_object_get_data (GTK_OBJECT (terrain_window),
                                     "data_parent");

  filter_name = gtk_object_get_data (GTK_OBJECT (filter_window), "data_title");
  filter_apply = gtk_object_get_data (GTK_OBJECT (filter_window),
                                      "data_filter_apply_func");

  filter_apply (filter_window, terrain);

  if (gtk_object_get_data (GTK_OBJECT (filter_window), "data_add_undo"))
    terrain_window_save_state (terrain_window, filter_name);

  terrain->modified = TRUE;

  scale = ((gfloat) TERRAIN_MIN_VIEW) / MAX (terrain->width, terrain->height);
  t_rgb_set_size (rgb, scale * terrain->width, scale * terrain->height);

  terrain_window_refresh_contours (terrain_window);
  terrain_window_queue_paint (terrain_window);

  gtk_widget_destroy (filter_window);
}


/*
 * filter_window_new: create an RGB, a reference preview of
 * terrain_window's terrain, a working copy of the preview, set up
 * the data pointers appropriately using gtk_object_set_data,
 * create connections from the hscales to the updated function, and
 * show the window
 */

void
filter_window_new (GtkWidget         *a_popup_menu_widget,
                   CreateWindowFunc   create_window_func,
                   FilterApply        filter_apply_func,
                   gchar            **hscales,
                   gboolean           add_undo_state)
{
  GtkWidget *terrain_window;
  GtkWidget *modal;

  terrain_window = lookup_toplevel (a_popup_menu_widget);

  modal = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_modal");
  if (modal == NULL)
    {
      gint       i;
      GtkWidget *filter_window;
      TRgb      *rgb;
      TTerrain  *terrain;
      TTerrain  *terrain_preview;

      terrain = gtk_object_get_data (GTK_OBJECT (terrain_window),
                                     "data_t_terrain");

      gtk_widget_set_sensitive (terrain_window, FALSE);
      filter_window = create_window_func (terrain);
      gtk_object_set_data (GTK_OBJECT (filter_window),
                           "data_filter_apply_func", filter_apply_func);
      for (i = 0; hscales[i] != NULL; i++)
        {
          GtkWidget *hscale;

          hscale = lookup_widget (filter_window, hscales[i]);
          gtk_signal_connect_object (
            GTK_OBJECT (GTK_RANGE (hscale)->adjustment),
            "value_changed",
            GTK_SIGNAL_FUNC (filter_window_update_preview),
            GTK_OBJECT (filter_window));
        }

      gtk_object_set_data (GTK_OBJECT (terrain_window), "data_modal",
                           filter_window);
      gtk_object_set_data_full (GTK_OBJECT (filter_window), "data_parent",
                                terrain_window, terrain_window_modal_destroy);

      terrain_preview = NULL;
      if (gtk_object_get_data (GTK_OBJECT (filter_window), "preview") != NULL)
        {
          terrain_preview = t_terrain_clone_preview (terrain);
          gtk_object_set_data_full (GTK_OBJECT (filter_window),
                                    "data_t_terrain_reference",
                                    terrain_preview,
                                    (GtkDestroyNotify) t_terrain_unref);
          terrain_preview = t_terrain_clone (terrain_preview);
          gtk_object_set_data_full (GTK_OBJECT (filter_window),
                                    "data_t_terrain",
                                    terrain_preview,
                                    (GtkDestroyNotify) t_terrain_unref);
          rgb = t_rgb_new (lookup_widget (filter_window, "preview"),
                           terrain_preview->width, terrain_preview->height);
          gtk_object_set_data_full (GTK_OBJECT (filter_window), "data_t_rgb",
                                    rgb, (GtkDestroyNotify) t_rgb_free);
        }

      if (terrain_preview != NULL &&
          gtk_object_get_data (GTK_OBJECT (filter_window), "center_x") != NULL)
        terrain_preview->view = T_VIEW_2D_PLANE;

      if (add_undo_state)
        gtk_object_set_data (GTK_OBJECT (filter_window),
                             "data_add_undo", filter_window);

      terrain_window_set_title (terrain_window, filter_window);
      filter_window_update_preview (filter_window);
      gtk_widget_show (filter_window);
    }
}
