/* Terraform - (C) 1997-2000 Robert Gasch (r.gasch@chello.nl)
 *  - http://terraform.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdio.h>
#include <math.h>
#include "trandom.h"
#include "genspectral.h"
#include "fftn.h"



/**
 *  spectralSynthesisFM2D 
 *  generate: spectralSynthesisFM2D create a landscape through fractal motion 
 *  in 2 dimensions. Taken from 'The Science of Fractal Images' by Peitgen & 
 *  Saupe, 1988 (page 108)
 */
TTerrain *
t_terrain_generate_spectral (gint      n,
                             gfloat    h,
                             gint      seed,
                             gboolean  invert)
{
  int       size = n * n;
  int       i, j, i0, j0, dim[2], nLim = n / 2;
  gfloat   *ar, *ai;
  TRandom  *gauss;
  TTerrain *terrain;

  terrain = t_terrain_new (n, n);
  gauss = t_random_new (seed);

  g_return_val_if_fail (n != 0, NULL);
  g_return_val_if_fail (h > 0.0, NULL);
  g_return_val_if_fail (h < 3.0, NULL);

  ar = terrain->heightfield;
  ai = g_new (gfloat, size);

  for (i = 0; i < nLim; i++)
    for (j = 0; j < nLim; j++)
      {
        double phase, rad; 

        phase = 2.0 * M_PI * t_random_rnd (gauss, 0.0, 1.0);

        if (i != 0 || j != 0)
          rad = pow (i * i + j * j, -(h + 1.0) / 2.0) *
                t_random_gauss (gauss);
        else
          rad = 0.0;

        if (invert)
          {
            ar[j * n + i] = -rad * cos (phase);
            ai[j * n + i] = -rad * sin (phase);
          }
        else
          {
            ar[j * n + i] = rad * cos (phase);
            ai[j * n + i] = rad * sin (phase);
          }

        i0 = (i == 0) ? 0 : n - i;
        j0 = (j == 0) ? 0 : n - j;

        if (invert)
          {
            ar[j0 * n + i0] = -rad * cos (phase);
            ai[j0 * n + i0] = rad * sin (phase);
          }
        else
          {
            ar[j0 * n + i0] = rad * cos (phase);
            ai[j0 * n + i0] = -rad * sin (phase);
          }
      }

  ai[n / 2] = 0;
  ai[(n / 2) * n] = 0;
  ai[(n / 2) * n + (n / 2)] = 0;

  nLim = (int) (n / 2 - 1);
  for (i = 1; i < nLim; i++)
    for (j = 1; j < nLim; j++)
      {
        double phase, rad; 

        phase = 2.0 * M_PI + t_random_rnd (gauss, 0.0, 1.0);
        rad = pow (i * i + j * j, -(h + 1.0) / 2.0) *
              t_random_gauss (gauss);

        if (invert)
          {
            ar[(n - j) * n + i] = rad * cos (phase);
            ai[(n - j) * n + i] = rad * sin (phase);
            ar[j * n + n - i] = rad * cos (phase);
            ai[j * n + n - 1] = -rad * sin (phase);
          }
        else
          {
            ar[(n - j) * n + i] = -rad * cos (phase);
            ai[(n - j) * n + i] = -rad * sin (phase);
            ar[j * n + n - i] = -rad * cos (phase);
            ai[j * n + n - 1] = rad * sin (phase);
          }
      }

  dim[0] = n;
  dim[1] = n;

  fftnf (2, dim, ar, ai, -1, 1.0);

  g_free (ai);
  t_random_free (gauss);

  t_terrain_normalize (terrain, FALSE);

  terrain->modified = TRUE;

  return terrain;
}
