/* Terraform - (C) 1997-2000 Robert Gasch (r.gasch@chello.nl)
 *  - http://terraform.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gnome.h>
#include <libgnomeui/gnome-window-icon.h>
#include "contour.h"
#include "tterrain.h"
#include "interface.h"
#include "support.h"
#include "support2.h"
#include "mainwindow.h"
#include "terrainwindow.h"
#include "trgb.h"
#include "tundo.h"
#include "colormaps.h"

static gchar *terrain_view_radio[] = {
  "terrain_2d_plane", "terrain_2d_contour",
  "terrain_3d_wire", "terrain_3d_height",
  "terrain_3d_light", NULL
};

static gchar *terrain_colormap_radio[] = {
  "terrain_land", "terrain_desert",
  "terrain_grayscale", "terrain_heat", NULL
};

static void
auto_rotate_func (gpointer terrain_window)
{
  TTerrain *terrain;

  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  if (terrain->view == T_VIEW_3D_WIRE)
    {
      TRgb *rgb;

      rgb = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_rgb");
      terrain->angle += 0.01;
      if (terrain->angle >= 2.0 * M_PI)
        terrain->angle -= 2.0 * M_PI;
      t_terrain_paint (terrain, rgb);
      t_rgb_soft_sync (rgb);
    }
}

static void
on_terrain_configure_event (GtkWidget         *drawing_area,
                            GdkEventConfigure *event,
                            GtkWidget         *terrain_window)
{
  TTerrain *terrain;
  gint     *size;

  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  /* Ignore move-only events.  All we care about are the resizes */
  size = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_last_size");
  if (size == NULL || size[0] != event->width || size[1] != event->height)
    {
      terrain_window_queue_paint (terrain_window);

      size = g_new (gint, 2);
      size[0] = event->width;
      size[1] = event->height;
      gtk_object_set_data_full (GTK_OBJECT (terrain_window),
                                "data_t_last_size", size,
                                (GtkDestroyNotify) g_free);
    }
}

GtkWidget *
terrain_window_new (GtkWidget *main_window,
                    TTerrain  *terrain)
{
  GtkWidget *terrain_window;
  GtkWidget *drawing_area;
  MainState *main_state;
  TUndo     *undo;

  terrain_window = create_terrain_window ();
  gtk_object_set_data (GTK_OBJECT (terrain_window), "data_parent", main_window);
  gnome_window_icon_set_from_default (GTK_WINDOW (terrain_window));

  terrain_window_set_terrain (terrain_window, terrain);

  main_state = gtk_object_get_data (GTK_OBJECT (main_window), "data_state");
  undo = t_undo_new (main_state->max_undo);
  gtk_object_set_data_full (GTK_OBJECT (terrain_window),
                            "data_t_undo", undo,
                            (GtkDestroyNotify) t_undo_free);

  drawing_area = lookup_widget (terrain_window, "terrain");
  gtk_signal_connect (GTK_OBJECT (drawing_area), "configure_event",
                      GTK_SIGNAL_FUNC (on_terrain_configure_event),
                      terrain_window);

  main_window_add_terrain_window (main_window, GTK_WIDGET (terrain_window));
  terrain_window_unpack_menu (terrain_window);
  gtk_widget_show (terrain_window);

  return terrain_window;
}


void
terrain_window_set_terrain (GtkWidget    *terrain_window,
                            TTerrain     *terrain)
{
  GtkWidget *aspectframe;
  gfloat     scale, min_width, min_height;
  TRgb      *rgb;

  /* Set the Terrain's filename if it has none. */
  if (terrain->filename[0] == '\0')
    {
      GtkWidget *main_window;

      main_window =
        gtk_object_get_data (GTK_OBJECT (terrain_window), "data_parent");

      g_free (terrain->filename);
      terrain->filename = main_window_filename_new (main_window);
      terrain->autogenned = TRUE;
    }

  scale = ((gfloat) TERRAIN_MIN_VIEW) / MAX (terrain->width, terrain->height);
  min_width = scale * terrain->width;
  min_height = scale * terrain->height;
  rgb = t_rgb_new (lookup_widget (terrain_window, "terrain"),
                   (gint) min_width, (gint) min_height);
  gtk_object_set_data_full (GTK_OBJECT (terrain_window),
                            "data_t_rgb", rgb,
                            (GtkDestroyNotify) t_rgb_free);

  aspectframe = lookup_widget (terrain_window, "aspectframe");
  gtk_aspect_frame_set (GTK_ASPECT_FRAME (aspectframe), 0.5, 0.5,
                        ((gfloat) terrain->height) / terrain->width, TRUE);

  t_terrain_ref (terrain);
  gtk_object_set_data_full (GTK_OBJECT (terrain_window),
                            "data_t_terrain", terrain,
                            (GtkDestroyNotify) t_terrain_unref);

  terrain_window_set_title (terrain_window, terrain_window);
  terrain_window_refresh_objects (terrain_window);

  terrain_window_queue_paint (terrain_window);
}


gchar *
terrain_window_get_title (GtkWidget *terrain_window)
{
  TTerrain *terrain;

  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");
  return g_strdup_printf ("%s%s\n",
                          terrain->filename,
                          terrain->modified ? " *" : "");
}

void
terrain_window_set_title (GtkWidget    *terrain_window,
                          GtkWidget    *toplevel_window)
{
  gchar        *title;
  gchar        *title_affix;

  /* Set the window's title. */
  title = gtk_object_get_data (GTK_OBJECT (toplevel_window), "data_title");
  if (title == NULL)
    {
      title = g_strdup (GTK_WINDOW (toplevel_window)->title);

      gtk_object_set_data_full (GTK_OBJECT (toplevel_window), "data_title",
        title, g_free);
    }

  title_affix = terrain_window_get_title (terrain_window);
  title = g_strdup_printf ("%s: %s", title, title_affix);
  gtk_window_set_title (GTK_WINDOW (toplevel_window), title);
  g_free (title);
  g_free (title_affix);
}

/*
 * Hard sync paints the terrain immediately, soft sync
 * will cause a gtk_queue_redraw_area.  Use hard sync
 * for animation to prevent dropped frames, use soft
 * sync otherwise.
 */

void
terrain_window_refresh (GtkWidget *terrain_window,
                        gboolean   hard_sync)
{
  TRgb         *rgb;
  TTerrain     *terrain;
  GtkWidget    *main_window;

  /* Set this window's title. */
  terrain_window_set_title (terrain_window, terrain_window);

  main_window =
    gtk_object_get_data (GTK_OBJECT (terrain_window), "data_parent");
  rgb =
    gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_rgb");
  terrain =
    gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  main_window_update_terrain_window (main_window, terrain_window);

  if (terrain->auto_rotate)
    widget_set_timer (terrain_window, 100, (GtkFunction) &auto_rotate_func);
  else
    widget_remove_timer (terrain_window);

  if (hard_sync)
    {
      t_terrain_paint (terrain, rgb);
      t_rgb_hard_sync (rgb);
    }
  else
    terrain_window_queue_paint (terrain_window);
}

/* Idle callback function to repaint the window */

static gboolean
terrain_window_paint_idle (gpointer user_data)
{
  GtkObject *terrain_window = GTK_OBJECT (user_data);
  TTerrain  *terrain;
  TRgb      *rgb;

  terrain = gtk_object_get_data (terrain_window, "data_t_terrain");
  rgb = gtk_object_get_data (terrain_window, "data_t_rgb");

  t_terrain_paint (terrain, rgb);
  t_rgb_ignore_expose (rgb, FALSE);
  t_rgb_soft_sync (rgb);
  gtk_object_remove_data (terrain_window, "data_paint_idle");

  return FALSE;
}

void
terrain_window_queue_paint (GtkWidget *terrain_window)
{
  guint *paint_idle;

  paint_idle =
    gtk_object_get_data (GTK_OBJECT (terrain_window), "data_paint_idle");

  if (paint_idle == NULL)
    {
      TRgb *rgb;

      rgb = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_rgb");
      t_rgb_ignore_expose (rgb, TRUE);

      paint_idle = g_new (guint, 1);
      *paint_idle = gtk_idle_add (terrain_window_paint_idle, terrain_window);
      gtk_object_set_data_full (GTK_OBJECT (terrain_window), "data_paint_idle",
                                paint_idle, (GtkDestroyNotify) g_free);
    }
}

/* terrain_window_terrain_refresh_contours: call this function if the terrain
   was updated.  It will recalculate contour lines. */
void
terrain_window_refresh_contours (GtkWidget *terrain_window)
{
  TTerrain *terrain;
  gint      view;

  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");
  view = get_radio_menu (terrain_window, terrain_view_radio);

  if (view == T_VIEW_2D_CONTOUR)
    t_terrain_contour_lines (terrain, terrain->contour_levels, 5);
}

/*
 * Add an undo state
 */

void
terrain_window_save_state (GtkWidget    *terrain_window,
                           gchar        *name)
{
  TUndo    *undo;
  TTerrain *terrain;

  undo = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_undo");
  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  t_undo_add_state (undo, terrain, name);
}

gboolean
terrain_window_exit_sequence (GtkWidget *terrain_window)
{
  GtkWidget *close_anyways;
  GtkWidget *parent;
  GtkWidget *modal;
  TTerrain *terrain;

  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");

  if (terrain->modified)
    {
      close_anyways = create_close_anyways_window ();
      gtk_object_set_data (GTK_OBJECT (close_anyways),
                           "data_parent", terrain_window);
      gtk_widget_show (close_anyways);

      return TRUE;
    }

  modal = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_modal");
  if (modal != NULL)
    gtk_widget_destroy (modal);

  parent = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_parent");
  main_window_remove_terrain_window (parent, terrain_window);

  return FALSE;
}

void
terrain_window_modal_destroy (gpointer user_data)
{
  gtk_widget_set_sensitive (GTK_WIDGET (user_data), TRUE);
  gtk_object_remove_data (GTK_OBJECT (user_data), "data_modal");
}

void
terrain_window_popup_destroy (gpointer user_data)
{
  gtk_object_remove_data (GTK_OBJECT (user_data), "data_popup");
}

void
terrain_window_refresh_objects (GtkWidget *terrain_window)
{
  TTerrain *terrain;
  TRgb     *rgb;
  GArray   *array;
  gint      i;

  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");
  rgb = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_rgb");
  array = terrain->objects;

  t_rgb_clear_objects (rgb);
  if (terrain->view == T_VIEW_2D_PLANE)
    {
      for (i = 0; i < array->len; i++)
        {
          TTerrainObject *object;

          object = &g_array_index (array, TTerrainObject, i);
          t_rgb_add_object (rgb, object->x, object->y, object->angle);
        }
    }

  gtk_widget_queue_draw (rgb->drawing_area);
}

void
terrain_window_pack_menu (GtkWidget    *terrain_window)
{
  TTerrain   *terrain;
  gint        view;
  TColormapTp colormap;
  gboolean    refresh;
  gboolean    auto_rotate;

  refresh = FALSE;

  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");
  view = get_radio_menu (terrain_window, terrain_view_radio);

  if (view == T_VIEW_2D_CONTOUR && terrain->view != T_VIEW_2D_CONTOUR)
    terrain_window_refresh_contours (terrain_window);

  if (view != -1 && terrain->view != view)
    {
      terrain->view = view;
      refresh = TRUE;
    }

  colormap = get_radio_menu (terrain_window, terrain_colormap_radio);
  if (colormap != -1 && terrain->colormap != colormap)
    {
      terrain->colormap = colormap;
      refresh = TRUE;
    }

  auto_rotate = get_boolean (terrain_window, "terrain_auto_rotate");
  if (auto_rotate != terrain->auto_rotate)
    {
      terrain->auto_rotate = auto_rotate;
      refresh = TRUE;
    }

  gtk_widget_set_sensitive (lookup_widget (terrain_window,
                                           "terrain_auto_rotate"),
                            terrain->view == T_VIEW_3D_WIRE);

  if (terrain->view == T_VIEW_2D_PLANE)
    gtk_widget_show (lookup_widget (terrain_window, "handlebox"));
  else
    gtk_widget_hide (lookup_widget (terrain_window, "handlebox"));

  if (refresh)
    {
      terrain_window_refresh_objects (terrain_window);
      terrain_window_refresh (terrain_window, FALSE);
    }
}

void
terrain_window_unpack_menu (GtkWidget    *terrain_window)
{
  TTerrain  *terrain;

  terrain = gtk_object_get_data (GTK_OBJECT (terrain_window), "data_t_terrain");
  if (terrain->view != T_VIEW_3D_WIRE)
    gtk_widget_set_sensitive (
      lookup_widget (terrain_window, "terrain_auto_rotate"), FALSE);

  set_boolean (terrain_window, "terrain_auto_rotate", terrain->auto_rotate);
  set_radio_menu (terrain_window, terrain_view_radio, terrain->view);
  set_radio_menu (terrain_window, terrain_colormap_radio, terrain->colormap);
}
