/* Terraform - (C) 1997-2000 Robert Gasch (r.gasch@chello.nl)
 *  - http://terraform.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "trgb.h"

#define SQR(x) ((x)*(x))

/* Set a region in the buffer as dirty. */
void
t_rgb_invalidate (TRgb *rgb,
                  int   x1,
                  int   y1,
                  int   x2,
                  int   y2)
{
  rgb->dirty_x1 = MIN (rgb->dirty_x1, x1);
  rgb->dirty_y1 = MIN (rgb->dirty_y1, y1);
  rgb->dirty_x2 = MAX (rgb->dirty_x2, x2);
  rgb->dirty_y2 = MAX (rgb->dirty_y2, y2);
}

/* Synchronise the screen every second. */

static gint
t_rgb_timeout (gpointer user_data)
{
  TRgb *rgb = (TRgb*) user_data;

  t_rgb_soft_sync (rgb);

  return TRUE;
}

static void
draw_selection (TRgb *rgb)
{
  if (!GTK_WIDGET_REALIZED (rgb->drawing_area))
    return;

  switch (rgb->selection_type)
    {
    case T_SELECTION_NONE:
      break;

    case T_SELECTION_RECTANGLE:
      gdk_draw_rectangle (rgb->drawing_area->window, rgb->xor_gc, FALSE,
                          rgb->selection_bounds.x,
                          rgb->selection_bounds.y,
                          rgb->selection_bounds.width,
                          rgb->selection_bounds.height);
      break;

    case T_SELECTION_ELLIPSE:
      gdk_draw_arc (rgb->drawing_area->window, rgb->xor_gc, FALSE,
                    rgb->selection_bounds.x,
                    rgb->selection_bounds.y,
                    rgb->selection_bounds.width,
                    rgb->selection_bounds.height,
                    0, 360 * 64 - 1);
      break;
    }
}

static void
draw_object (TRgb *rgb,
             gint  item)
{
  gfloat x, y, *data;

  /* FIXME: Show angle */
  data = &g_array_index (rgb->object_array, gfloat, item * 3);
  x = data[0] * rgb->width;
  y = data[1] * rgb->height;
  gdk_draw_line (rgb->drawing_area->window, rgb->xor_gc, x - 4, y - 4, x + 4, y + 4);
  gdk_draw_line (rgb->drawing_area->window, rgb->xor_gc, x + 4, y - 4, x - 4, y + 4);
}

static void
draw_objects (TRgb *rgb)
{
  gint i, size;

  size = rgb->object_array->len / 3;
  for (i = 0; i < size; i++)
    draw_object (rgb, i);
}

static void
on_expose_event (GtkWidget      *drawing_area,
                 GdkEventExpose *event,
                 TRgb           *rgb)
{
  GdkGC        *gc;
  GdkWindow    *window;
  gint          x1, y1, x2, y2;

  if (rgb->ignore_expose)
    return;

  x1 = MIN (MAX (0, event->area.x), rgb->width - 1);
  y1 = MIN (MAX (0, event->area.y), rgb->height - 1);
  x2 = MIN (MAX (0, event->area.x + event->area.width - 1), rgb->width - 1);
  y2 = MIN (MAX (0, event->area.y + event->area.height - 1), rgb->height - 1);

  gc = rgb->drawing_area->style->black_gc;
  window = rgb->drawing_area->window;
  gdk_draw_rgb_image_dithalign (window, gc, x1, y1,
                                x2 - x1 + 1, y2 - y1 + 1,
                                GDK_RGB_DITHER_NORMAL,
                                &rgb->data[(y1 * rgb->width + x1) * 3],
                                rgb->width * 3, x1, y1);

  gdk_gc_set_clip_rectangle (rgb->xor_gc, &event->area);
  draw_selection (rgb);
  draw_objects (rgb);
  gdk_gc_set_clip_mask (rgb->xor_gc, NULL);
}

static void
on_configure_event (GtkWidget         *drawing_area,
                    GdkEventConfigure *event,
                    TRgb              *rgb)
{
  if (event->width == rgb->width && event->height == rgb->height)
    return;

  g_free (rgb->data);
  rgb->width = event->width;
  rgb->height = event->height;
  rgb->data = g_new0 (guchar, rgb->width * rgb->height * 3);

  rgb->dirty_x1 = 0;
  rgb->dirty_y1 = 0;
  rgb->dirty_x2 = rgb->width - 1;
  rgb->dirty_y2 = rgb->height - 1;
}


static void
drawing_area_realize (GtkWidget *drawing_area,
                      TRgb      *rgb)
{
  GtkWidget   *widget;
  GdkGCValues  values;
  gint         mask;

  widget = rgb->drawing_area;
  values.foreground = (widget->style->white.pixel == 0) ?
                      widget->style->black : widget->style->white;
  values.function = GDK_XOR;
  values.subwindow_mode = GDK_INCLUDE_INFERIORS;
  mask = GDK_GC_FOREGROUND | GDK_GC_FUNCTION | GDK_GC_SUBWINDOW;
  rgb->xor_gc = gdk_gc_new_with_values (widget->window, &values, mask);
}

static void
drawing_area_unrealize (GtkWidget *drawing_area,
                        TRgb      *rgb)
{
  gdk_gc_destroy (rgb->xor_gc);
  rgb->xor_gc = NULL;
}

static void
drawing_area_destroy (GtkWidget *drawing_area,
                      TRgb      *rgb)
{
  gtk_signal_disconnect (GTK_OBJECT (rgb->drawing_area), rgb->expose_id);
  gtk_signal_disconnect (GTK_OBJECT (rgb->drawing_area), rgb->configure_id);
  gtk_signal_disconnect (GTK_OBJECT (rgb->drawing_area), rgb->realize_id);
  gtk_signal_disconnect (GTK_OBJECT (rgb->drawing_area), rgb->unrealize_id);
  gtk_signal_disconnect (GTK_OBJECT (rgb->drawing_area), rgb->destroy_id);

  rgb->expose_id = 0;
  rgb->configure_id = 0;
  rgb->realize_id = 0;
  rgb->unrealize_id = 0;
  rgb->destroy_id = 0;

  rgb->drawing_area = NULL;
}

/*
 * Create a new RGB double buffer.  The RGB buffer will repaint itself
 * using the provided drawing_area.
 */

TRgb *
t_rgb_new (GtkWidget *drawing_area,
           gint       width,
           gint       height)
{
  TRgb        *rgb;

  rgb = g_new0 (TRgb, 1);
  rgb->drawing_area = drawing_area;
  gtk_object_ref (GTK_OBJECT (drawing_area));
  gtk_object_sink (GTK_OBJECT (drawing_area));

  rgb->object_array = g_array_new (FALSE, FALSE, sizeof (gfloat));
  rgb->xor_gc = NULL;
  rgb->selection_type = T_SELECTION_NONE;
  rgb->width = width;
  rgb->height = height;
  rgb->data = g_new0 (guchar, width * height * 3);
  rgb->dirty_x1 = width + 1;
  rgb->dirty_y1 = height + 1;
  rgb->dirty_x2 = -1;
  rgb->dirty_y2 = -1;
  rgb->sync_timeout = gtk_timeout_add (1000, t_rgb_timeout, rgb);
  rgb->fore[0] = 0xff; rgb->fore[1] = 0xff; rgb->fore[2] = 0xff;
  rgb->back[0] = 0; rgb->back[1] = 0; rgb->back[2] = 0;

  rgb->expose_id =
    gtk_signal_connect (GTK_OBJECT (drawing_area), "expose_event",
                        GTK_SIGNAL_FUNC (on_expose_event), rgb);

  rgb->configure_id =
    gtk_signal_connect (GTK_OBJECT (drawing_area), "configure_event",
                        GTK_SIGNAL_FUNC (on_configure_event), rgb);

  rgb->realize_id =
    gtk_signal_connect (GTK_OBJECT (drawing_area), "realize",
                        GTK_SIGNAL_FUNC (drawing_area_realize), rgb);

  rgb->unrealize_id =
    gtk_signal_connect (GTK_OBJECT (drawing_area), "unrealize",
                        GTK_SIGNAL_FUNC (drawing_area_unrealize), rgb);

  rgb->destroy_id =
    gtk_signal_connect (GTK_OBJECT (drawing_area), "destroy",
                        GTK_SIGNAL_FUNC (drawing_area_destroy), rgb);

  if (GTK_WIDGET_REALIZED (drawing_area))
    drawing_area_realize (drawing_area, rgb);

  if (width > 0 && height > 0)
    gtk_drawing_area_size (GTK_DRAWING_AREA (drawing_area), width, height);

  return rgb;
}

void
t_rgb_free (TRgb *rgb)
{
  if (rgb->drawing_area != NULL)
    {
      gtk_signal_disconnect (GTK_OBJECT (rgb->drawing_area), rgb->expose_id);
      gtk_signal_disconnect (GTK_OBJECT (rgb->drawing_area), rgb->configure_id);
      gtk_signal_disconnect (GTK_OBJECT (rgb->drawing_area), rgb->realize_id);
      gtk_signal_disconnect (GTK_OBJECT (rgb->drawing_area), rgb->unrealize_id);
      gtk_signal_disconnect (GTK_OBJECT (rgb->drawing_area), rgb->destroy_id);

      gtk_object_unref (GTK_OBJECT (rgb->drawing_area));
    }

  g_array_free (rgb->object_array, TRUE);
  gtk_timeout_remove (rgb->sync_timeout);
  g_free (rgb->data);
  g_free (rgb);
}

void
t_rgb_set_size (TRgb     *rgb,
                gint      width,
                gint      height)
{
  if (rgb->width == width && rgb->height == height)
    return;

  rgb->width = width;
  rgb->height = height;
  g_free (rgb->data);
  rgb->data = g_new (guchar, width * height * 3);

  gtk_drawing_area_size (GTK_DRAWING_AREA (rgb->drawing_area),
                         width, height);
  gtk_widget_unmap (rgb->drawing_area);
  gtk_widget_map (rgb->drawing_area);
}

void
t_rgb_set_fore (TRgb *rgb,
                gint  red,
                gint  green,
                gint  blue)
{
  rgb->fore[0] = red;
  rgb->fore[1] = green;
  rgb->fore[2] = blue;
}

void
t_rgb_set_back (TRgb *rgb,
                gint  red,
                gint  green,
                gint  blue)
{
  rgb->back[0] = red;
  rgb->back[1] = green;
  rgb->back[2] = blue;
}

void
t_rgb_draw_raster (TRgb   *rgb,
                   gint    y,
                   guchar *data)
{
  if (y >= 0 && y < rgb->height)
    {
      memcpy (&rgb->data[y * rgb->width * 3], data, rgb->width * 3);
      t_rgb_invalidate (rgb, 0, y, rgb->width, y);
    }
}

void
t_rgb_draw_pel (TRgb     *rgb,
                gint      x,
                gint      y)
{
  if (x >= 0 && y >= 0 && x < rgb->width && y < rgb->height)
    {
      guchar *pos;

      pos = &rgb->data[(y * rgb->width + x) * 3];
      pos[0] = rgb->fore[0];
      pos[1] = rgb->fore[1];
      pos[2] = rgb->fore[2];

      t_rgb_invalidate (rgb, x, y, x, y);
    }
}

void
t_rgb_get_pel (TRgb *rgb,
               gint  x,
               gint  y,
               gint *r,
               gint *g,
               gint *b)
{
  if (x >= 0 && y >= 0 && x < rgb->width && y < rgb->height)
    {
      guchar *pos;

      pos = &rgb->data[(y * rgb->width + x) * 3];
      *r = pos[0];
      *g = pos[1];
      *b = pos[2];
    }
  else
    {
      *r = 0;
      *g = 0;
      *b = 0; 
    }
}

void
t_rgb_draw_line (TRgb     *rgb,
                 gint      x1,
                 gint      y1,
                 gint      x2,
                 gint      y2)
{
  int x, y;
  int count;
  int delta_x, delta_y;
  int dir_x, dir_y;

  delta_x = x2 - x1;
  if (delta_x < 0) delta_x = -delta_x, dir_x = -1; else dir_x = 1;

  delta_y = y2 - y1;
  if (delta_y < 0) delta_y = -delta_y, dir_y = -1; else dir_y = 1;

  x = x1; 
  y = y1;
  count = 0;
  if (delta_x > delta_y)
    {
      for (; x != x2 + dir_x; x += dir_x)
        {
          count += delta_y;
          if (count > delta_x) count -= delta_x, y += dir_y;

          if (x >= 0 && y >= 0 && x < rgb->width && y < rgb->height)
            {
              guchar *pos;

              pos = &rgb->data[(y * rgb->width + x) * 3];
              pos[0] = rgb->fore[0];
              pos[1] = rgb->fore[1];
              pos[2] = rgb->fore[2];
            }
        }
    }
  else
    {
      for (; y != y2 + dir_y; y += dir_y)
        {
          count += delta_x;
          if (count > delta_y) count -= delta_y, x += dir_x;

          if (x >= 0 && y >= 0 && x < rgb->width && y < rgb->height)
            {
              guchar *pos;

              pos = &rgb->data[(y * rgb->width + x) * 3];
              pos[0] = rgb->fore[0];
              pos[1] = rgb->fore[1];
              pos[2] = rgb->fore[2];
            }
        }
    }

  t_rgb_invalidate (rgb, x1, y1, x2, y2);
}

void
t_rgb_draw_line_add (TRgb     *rgb,
                     gint      x1,
                     gint      y1,
                     gint      x2,
                     gint      y2)
{
  int x, y;
  int count;
  int delta_x, delta_y;
  int dir_x, dir_y;

  delta_x = x2 - x1;
  if (delta_x < 0) delta_x = -delta_x, dir_x = -1; else dir_x = 1;

  delta_y = y2 - y1;
  if (delta_y < 0) delta_y = -delta_y, dir_y = -1; else dir_y = 1;

  x = x1; 
  y = y1;
  count = 0;
  if (delta_x > delta_y)
    {
      for (; x != x2 + dir_x; x += dir_x)
        {
          count += delta_y;
          if (count > delta_x) count -= delta_x, y += dir_y;

          if (x >= 0 && y >= 0 && x < rgb->width && y < rgb->height)
            {
              guchar *pos;

              pos = &rgb->data[(y * rgb->width + x) * 3];
              pos[0] = MIN (pos[0] + rgb->fore[0], 255);
              pos[1] = MIN (pos[1] + rgb->fore[1], 255);
              pos[2] = MIN (pos[2] + rgb->fore[2], 255);
            }
        }
    }
  else
    {
      for (; y != y2 + dir_y; y += dir_y)
        {
          count += delta_x;
          if (count > delta_y) count -= delta_y, x += dir_x;

          if (x >= 0 && y >= 0 && x < rgb->width && y < rgb->height)
            {
              guchar *pos;

              pos = &rgb->data[(y * rgb->width + x) * 3];
              pos[0] = MIN (pos[0] + rgb->fore[0], 255);
              pos[1] = MIN (pos[1] + rgb->fore[1], 255);
              pos[2] = MIN (pos[2] + rgb->fore[2], 255);
            }
        }
    }

  t_rgb_invalidate (rgb, x1, y1, x2, y2);
}

void
t_rgb_draw_arrow (TRgb      *rgb,
                  gint       x1,
                  gint       y1,
                  gint       x2,
                  gint       y2)
{
  const gfloat arrow_angle = 0.25 * M_PI;
  const gfloat arrow_length = 8.0;

  t_rgb_draw_line (rgb, x1, y1, x2, y2);
  if (x1 != x2 || y1 != y2)
    {
      gint   x3, y3, x4, y4;
      gfloat angle, dist;

      dist = sqrt (SQR (x2 - x1) + SQR (y2 - y1));
      angle = acos ((y1 - y2) / dist); /* 0 -> PI */
      if (x2 - x1 < 0)
        angle = -angle;

      dist = MIN (dist, arrow_length);

      x3 = rint (x2 + cos (angle + arrow_angle) * dist);
      x4 = rint (x2 - cos (angle - arrow_angle) * dist);
      y3 = rint (y2 + sin (angle + arrow_angle) * dist);
      y4 = rint (y2 - sin (angle - arrow_angle) * dist);

      t_rgb_draw_line (rgb, x2, y2, x3, y3);
      t_rgb_draw_line (rgb, x2, y2, x4, y4);
    }
}

void
t_rgb_draw_rect (TRgb      *rgb,
                 gint       x1,
                 gint       y1,
                 gint       x2,
                 gint       y2)
{
  t_rgb_draw_line (rgb, x1, y1, x2, y1);
  t_rgb_draw_line (rgb, x2, y1, x2, y2);
  t_rgb_draw_line (rgb, x2, y2, x1, y2);
  t_rgb_draw_line (rgb, x1, y2, x1, y1);
}

void
t_rgb_draw_poly (TRgb     *rgb,
                 gint      point_count,
                 GdkPoint *points)
{
  int i;

  for (i = 0; i < point_count - 1; i++)
    t_rgb_draw_line (rgb, points[i].x, points[i].y,
                     points[i + 1].x, points[i + 1].y);

  t_rgb_draw_line (rgb, points[i].x, points[i].y,
                   points[0].x, points[0].y);
}

/*
 * Fill the pixmap with TV static.  A cheezy LFSR is used to generate
 * pseudorandom noise.
 */

void
t_rgb_render_static (TRgb *rgb)
{
  gint           i, size;
  static guint16 lfsr;

  if (lfsr == 0)
    lfsr = time (NULL);

  srand (lfsr);
  lfsr = rand ();
  if (lfsr == 0) lfsr++;

  size = rgb->width * rgb->height * 3;
  for (i = 0; i < size; i += 3)
    {
      lfsr = ((lfsr >> 1) | (lfsr << 15)) ^ (lfsr & 0x7f);

      rgb->data[i + 0] = lfsr;
      rgb->data[i + 1] = lfsr;
      rgb->data[i + 2] = lfsr;
    }

  t_rgb_invalidate (rgb, 0, 0, rgb->width - 1, rgb->height - 1);
}

void
t_rgb_clear (TRgb *rgb)
{
  gint     i, length;

  length = rgb->width * rgb->height * 3;

  for (i = 0; i < length; i += 3)
    {
      rgb->data[i + 0] = rgb->back[0];
      rgb->data[i + 1] = rgb->back[1];
      rgb->data[i + 2] = rgb->back[2];
    }

  t_rgb_invalidate (rgb, 0, 0, rgb->width - 1, rgb->height - 1);
}

static void
t_rgb_draw_horiz_line (TRgb *rgb,
                       gint  x1,
                       gint  y,
                       gint  x2)
{
  guchar *pos;

  if (x2 < 0 || x1 >= rgb->width || y < 0 || y >= rgb->height)
    return;

  x1 = MAX (x1, 0);
  x2 = MIN (x2, rgb->width - 1);

  pos = &rgb->data[(y * rgb->width + x1) * 3];

  for (; x1 <= x2; x1++)
    {
      pos[0] = rgb->fore[0];
      pos[1] = rgb->fore[1];
      pos[2] = rgb->fore[2];

      pos += 3;
    }
}

int
point_comparison (const void *point_1, const void *point_2)
{
  return ((GdkPoint*) point_1)->y > ((GdkPoint*) point_2)->y;
}

void
t_rgb_fill_triangle (TRgb     *rgb,
                     GdkPoint *points)
{
  GdkPoint pts[3];
  gint     count_1, count_2;
  gint     max_1, max_2;
  gint     inc_1, inc_2;
  gint     x_1, x_2;
  gint     dir_1, dir_2;
  gint     y;

  if (points[0].y == points[1].y && points[1].y == points[2].y)
    {
      t_rgb_draw_horiz_line (
        rgb, MIN (points[0].x, MIN (points[1].x, points[2].x)),
             points[0].y,
             MAX (points[0].x, MAX (points[1].x, points[2].x)));

      return;
    }

  /* Copy and sort points by y-coordinate. */
  memcpy (pts, points, sizeof (GdkPoint) * 3);
  qsort (pts, 3, sizeof (GdkPoint), &point_comparison);

  x_1 = x_2 = pts[0].x;
  dir_1 = (pts[0].x < pts[2].x) ? 1 : -1;
  dir_2 = (pts[0].x < pts[1].x) ? 1 : -1;
  max_1 = abs (pts[0].y - pts[2].y);
  max_2 = abs (pts[0].y - pts[1].y);
  inc_1 = abs (pts[0].x - pts[2].x);
  count_1 = 0;
  inc_2 = abs (pts[0].x - pts[1].x);
  count_2 = 0;
  for (y = pts[0].y; y <= pts[2].y; y++)
    {
      if (y == pts[1].y)
        {
          x_2 = pts[1].x;
          max_2 = abs (pts[1].y - pts[2].y);
          inc_2 = abs (pts[1].x - pts[2].x);
          count_2 = 0;
          dir_2 = (pts[1].x < pts[2].x) ? 1 : -1;
        }

      t_rgb_draw_horiz_line (rgb, MIN (x_1, x_2), y, MAX (x_1, x_2));

      count_1 += inc_1;
      if (max_1 > 0)
        while (count_1 >= max_1)
          {
            count_1 -= max_1;
            x_1 += dir_1;
          }

      count_2 += inc_2;
      if (max_2 > 0)
        while (count_2 >= max_2)
          {
            count_2 -= max_2;
            x_2 += dir_2;
          }
    }

  /* FIXME: doesn't call t_rgb_invalidate */
}

void
t_rgb_fill_rect (TRgb *rgb,
                 gint  x1,
                 gint  y1,
                 gint  x2,
                 gint  y2)
{
  gint y;

  if (x2 < 0 || y2 < 0 || x1 >= rgb->width || y1 >= rgb->height) return;
  x1 = MAX (x1, 0);
  y1 = MAX (y1, 0);
  x2 = MIN (x2, rgb->width - 1);
  y2 = MIN (y2, rgb->height - 1);

  g_return_if_fail (x1 <= x2 && y1 <= y2);

  for (y = y1; y <= y2; y++)
    {
      gint    x;
      guchar *pos;

      pos = &rgb->data[(y * rgb->width + x1) * 3];
      for (x = x1; x <= x2; x++)
        {
          pos[0] = rgb->fore[0];
          pos[1] = rgb->fore[1];
          pos[2] = rgb->fore[2];

          pos += 3;
        }
    }

  t_rgb_invalidate (rgb, x1, y1, x2, y2);
}

TSelectionType
t_rgb_get_selection (TRgb      *rgb,
                     gfloat    *x1,
                     gfloat    *y1,
                     gfloat    *x2,
                     gfloat    *y2)
{
  if (rgb->selection_type == T_SELECTION_NONE)
    return rgb->selection_type;

  *x1 = ((gfloat) rgb->selection_bounds.x) / rgb->width;
  *y1 = ((gfloat) rgb->selection_bounds.y) / rgb->height;
  *x2 = *x1 + ((gfloat) rgb->selection_bounds.width) / rgb->width;
  *y2 = *y1 + ((gfloat) rgb->selection_bounds.height) / rgb->height;

  return rgb->selection_type;
}

void
t_rgb_set_selection (TRgb           *rgb,
                     TSelectionType  select_type,
                     gint           x1,
                     gint           y1,
                     gint           x2,
                     gint           y2)
{
  draw_selection (rgb);

  if (x2 < x1 || y2 < y1)
    select_type = T_SELECTION_NONE;

  rgb->selection_type = select_type;
  rgb->selection_bounds.x = x1;
  rgb->selection_bounds.y = y1;
  rgb->selection_bounds.width = x2 - x1 + 1;
  rgb->selection_bounds.height = y2 - y1 + 1;
  draw_selection (rgb);
}

gint
t_rgb_add_object (TRgb      *rgb,
                  gfloat     x,
                  gfloat     y,
                  gfloat     angle)
{
  gfloat array[3];

  array[0] = x;
  array[1] = y;
  array[2] = angle;
  g_array_append_vals (rgb->object_array, &array, 3);

  return rgb->object_array->len / 3 - 1;
}

void
t_rgb_modify_object (TRgb      *rgb,
                     gint       item,
                     gfloat     x,
                     gfloat     y,
                     gfloat     angle)
{
  gint index;

  draw_object (rgb, item);
  index = item * 3;
  g_array_index (rgb->object_array, gfloat, index) = x;
  g_array_index (rgb->object_array, gfloat, index + 1) = y;
  g_array_index (rgb->object_array, gfloat, index + 2) = angle;
  draw_object (rgb, item);
}

void
t_rgb_get_object (TRgb      *rgb,
                  gint       item,
                  gfloat    *x,
                  gfloat    *y,
                  gfloat    *angle)
{
  gfloat *obj;

  obj = &g_array_index (rgb->object_array, gfloat, item * 3);
  *x = obj[0];
  *y = obj[1];
  *angle = obj[2];
}

void
t_rgb_clear_objects (TRgb *rgb)
{
  g_array_free (rgb->object_array, TRUE);
  rgb->object_array = g_array_new (FALSE, FALSE, sizeof (gfloat));
}

/* t_rgb_soft_sync adds the dirty region to Gtk+'s exposure queue. */

void
t_rgb_soft_sync (TRgb *rgb)
{
  if (rgb->dirty_x1 <= rgb->dirty_x2 &&
      rgb->dirty_y1 <= rgb->dirty_y2)
    {
      gtk_widget_queue_draw_area (rgb->drawing_area,
                                  rgb->dirty_x1, rgb->dirty_y1,
                                  rgb->dirty_x2 - rgb->dirty_x1 + 1,
                                  rgb->dirty_y2 - rgb->dirty_y1 + 1);

      rgb->dirty_x1 = rgb->width + 1;
      rgb->dirty_y1 = rgb->height + 1;
      rgb->dirty_x2 = -1;
      rgb->dirty_y2 = -1;
    }
}

/*
 * t_rgb_hard_sync paints the dirty region immediately without using
 * Gtk+'s exposure queue.  Generally you want to use soft sync unless
 * you are updating the window quickly (i.e. via a mouse drag).  In
 * that case you want to use hard sync to prevent frames from being dropped.
 */

void
t_rgb_hard_sync (TRgb *rgb)
{
  if (!GTK_WIDGET_REALIZED (rgb->drawing_area))
    return;

  if (rgb->dirty_x1 <= rgb->dirty_x2 &&
      rgb->dirty_y1 <= rgb->dirty_y2)
    {
      GdkEventExpose event;

      event.area.x = rgb->dirty_x1;
      event.area.y = rgb->dirty_y1;
      event.area.width = rgb->dirty_x2 - rgb->dirty_x1 + 1;
      event.area.height = rgb->dirty_y2 - rgb->dirty_y1 + 1;

      on_expose_event (rgb->drawing_area, &event, rgb);

      rgb->dirty_x1 = rgb->width + 1;
      rgb->dirty_y1 = rgb->height + 1;
      rgb->dirty_x2 = -1;
      rgb->dirty_y2 = -1;
    }
}

GtkWidget *
t_rgb_get_drawing_area (TRgb *rgb)
{
  return rgb->drawing_area;
}

void
t_rgb_ignore_expose (TRgb      *rgb,
                     gboolean   ignore)
{
  rgb->ignore_expose = ignore;
}
