/* Terraform - (C) 1997-2000 Robert Gasch (r.gasch@chello.nl)
 *  - http://terraform.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <math.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gnome-xml/tree.h>
#include <gnome-xml/parser.h>

#include "tterrain.h"
#include "support.h"
#include "support2.h"
#include "base64.h"
#include "xmlsupport.h"
#include "terrainwindow.h"
#include "filterwindow.h"
#include "contour.h"

#define CLIP(x,m,M) MAX(MIN(x,M),m)

void
t_terrain_object_clear (TTerrainObject *object)
{
  g_free (object->name);
  memset (object, 0, sizeof (TTerrainObject));
}

static gint
g_list_string_index (GList *list,
                     gchar *string)
{
  gint i;

  g_return_val_if_fail (list != NULL, -1);
  g_return_val_if_fail (string != NULL, -1);

  for (i = 0, list = g_list_first (list); list != NULL; i++, list = list->next)
    if (!strcmp (list->data, string))
      return i;

  return -1;
}

TTerrain *
t_terrain_new (gint width,
               gint height)
{
  TTerrain *terrain;

  terrain = g_new (TTerrain, 1);
  terrain->ref_count = 1;
  terrain->filename = g_strdup ("");
  terrain->modified = FALSE;
  terrain->autogenned = FALSE;
  terrain->width = width;
  terrain->height = height;
  terrain->heightfield = g_new (gfloat, width * height);
  terrain->selection = NULL;
  terrain->camera_x = 0.5;
  terrain->camera_y = 2.5;
  terrain->camera_z = 1.25;
  terrain->lookat_x = 0.5;
  terrain->lookat_y = 0.0;
  terrain->lookat_z = 0.5;
  terrain->observe_sealevel = TRUE;
  terrain->elevation_offset = 0.1;
  terrain->smooth_terrain = TRUE;
  terrain->clouds = TRUE;
  terrain->fog = FALSE;
  terrain->filled_sea = TRUE;
  terrain->time_of_day = 12.0;
  terrain->north_direction = 0.0;
  terrain->water_clarity = 0.65;
  terrain->render_width = 400;
  terrain->y_scale_factor = 0.33;
  terrain->sealevel = 0.33;
  terrain->wireframe_resolution = 10;
  terrain->theme_file = g_strdup ("earth_green.pov");
  terrain->lighting_level = 0;
  terrain->contour_levels = 10;

  terrain->objects = g_array_new (FALSE, FALSE, sizeof (TTerrainObject));
  terrain->contour_lines = NULL;

  terrain->view = T_VIEW_2D_PLANE;
  terrain->colormap = T_COLORMAP_LAND;
  terrain->angle = 0.0;
  terrain->elevation = 0.25;
  terrain->auto_rotate = FALSE;

  return terrain;
}

void
t_terrain_free (TTerrain *terrain)
{
  t_terrain_remove_objects (terrain);
  g_array_free (terrain->objects, TRUE);

  t_terrain_contour_list_free (terrain);
  g_free (terrain->heightfield);
  g_free (terrain->selection);
  g_free (terrain->filename);
  g_free (terrain->theme_file);
}

TTerrain *
t_terrain_clone (TTerrain *terrain)
{
  TTerrain *out;

  out = g_memdup (terrain, sizeof (TTerrain));
  out->filename = g_strdup (terrain->filename);
  out->theme_file = g_strdup (terrain->theme_file);
  out->heightfield =
    g_memdup (terrain->heightfield,
              sizeof (gfloat) * terrain->width * terrain->height);
  if (terrain->selection != NULL)
    out->selection =
      g_memdup (terrain->selection,
                sizeof (gfloat) * terrain->width * terrain->height);
  out->contour_lines = NULL;
  out->objects = g_array_new (FALSE, FALSE, sizeof (TTerrainObject));

  return out;
}

TTerrain *
t_terrain_clone_resize (TTerrain *terrain,
                        gint      width,
                        gint      height)
{
  TTerrain *out;
  gint      y;
  gint      pos;

  out = g_memdup (terrain, sizeof (TTerrain));
  out->filename = g_strdup (terrain->filename);
  out->theme_file = g_strdup (terrain->theme_file);
  out->width = width;
  out->height = height;

  out->heightfield =
    g_new (gfloat, sizeof (gfloat) * out->width * out->height);
  if (terrain->selection != NULL)
    out->selection =
      g_new (gfloat, sizeof (gfloat) * out->width * out->height);
  out->contour_lines = NULL;
  out->objects = g_array_new (FALSE, FALSE, sizeof (TTerrainObject));

  pos = 0;
  for (y = 0; y < out->height; y++)
    {
      gint x;
      gint count;
      gint prev_pos;

      prev_pos = y * terrain->height / out->height * terrain->width;
      count = 0;
      for (x = 0; x < out->width; x++)
        {
          out->heightfield[pos] = terrain->heightfield[prev_pos];
          if (out->selection != NULL)
            out->selection[pos] = terrain->selection[prev_pos];

          pos++;
          count += terrain->width;
          while (count >= out->width)
            {
              count -= out->width;
              prev_pos++;
            }
        }
    }

  return out;
}

void
t_terrain_set_size (TTerrain *terrain,
                    gint      width,
                    gint      height)
{
  if (terrain->width != width || terrain->height != height)
    {
      terrain->width = width;
      terrain->height = height;
      g_free (terrain->heightfield);
      g_free (terrain->selection);
      terrain->heightfield = g_new0 (gfloat, width * height);
      terrain->selection = NULL;
    }
}

TTerrain *
t_terrain_clone_preview (TTerrain *terrain)
{
  gint      maximum;
  TTerrain *clone;

  maximum = MAX (terrain->width, terrain->height);

  clone = t_terrain_clone_resize (terrain,
                                  MAX (1, terrain->width * 100 / maximum),
                                  MAX (1, terrain->height * 100 / maximum));

  clone->wireframe_resolution = 5;

  return clone;
}

void
t_terrain_copy_heightfield (TTerrain *from,
                            TTerrain *to)
{
  gint i, size;

  g_return_if_fail (from->width == to->width &&
                    from->height == to->height);

  size = to->width * to->height;
  for (i = 0; i < size; i++)
    to->heightfield[i] = from->heightfield[i];
  to->sealevel = from->sealevel;
}

void
t_terrain_copy_heightfield_and_selection (TTerrain   *from,
                                          TTerrain   *to)
{
  gint i, size;

  g_return_if_fail (from->width == to->width &&
                    from->height == to->height);

  size = to->width * to->height;
  for (i = 0; i < size; i++)
    to->heightfield[i] = from->heightfield[i];

  if (from->selection == NULL)
    {
      g_free (to->selection);
      to->selection = NULL;
    }
  else
    {
      if (to->selection == NULL)
        to->selection = g_new (gfloat, from->width * from->height);
      for (i = 0; i < size; i++)
        to->selection[i] = from->selection[i];
    }

  to->sealevel = from->sealevel;
}

void
t_terrain_normalize (TTerrain *terrain,
                     gboolean  never_grow)
{
  int     i, last;
  gfloat  min, max;
  gfloat *data;

  data = terrain->heightfield;
  last = terrain->width * terrain->height;

  /*
   * When never_grow is true, normalize assumes that if the minimum
   * and maximum value range of the terrain is smaller than [0.0, 1.0],
   * the user wants it that way.
   */

  if (never_grow)
    min = 0.0, max = 1.0;
  else
    min = max = data[0];

  for (i = 0; i < last; i++)
    {
      if (data[i] < min) min = data[i];
      if (data[i] > max) max = data[i];
    }

  if (fabs (max - min) > 0.0001)
    {
      gfloat k;

      k = 1.0 / (max - min);
      for (i = 0; i < last; i++)
        data[i] = (data[i] - min) * k;

      if (never_grow)
        terrain->sealevel =
          MIN (MAX ((terrain->sealevel - min) * k, 0.0), 1.0);
    }
}

void
t_terrain_crop (TTerrain *terrain,
                gint      x1,
                gint      y1,
                gint      x2,
                gint      y2)
{ 
  int from_x, from_y;
  int to_x, to_y;

  g_return_if_fail (x1 >= 0 && y1 >= 0 &&
                    x2 < terrain->width && y2 < terrain->height);
  g_return_if_fail (x1 <= x2 && y1 <= y2);

  for (from_y = y1, to_y = 0; from_y <= y2; from_y++, to_y++)
    {
      int from_pos, to_pos;

      from_pos = from_y * terrain->width + x1;
      to_pos = to_y * (x2 - x1 + 1);
      for (from_x = x1, to_x = 0; from_x <= x2; from_x++, to_x++)
        {
          terrain->heightfield[to_pos] = terrain->heightfield[from_pos];
          if (terrain->selection != NULL)
            terrain->selection[to_pos] = terrain->selection[from_pos];

          from_pos++;
          to_pos++;
        }
    }

  terrain->width = x2 - x1 + 1;
  terrain->height = y2 - y1 + 1;
  terrain->heightfield = g_renew (gfloat, terrain->heightfield,
                                  terrain->width * terrain->height);
  if (terrain->selection != NULL)
    terrain->selection = g_renew (gfloat, terrain->selection,
                                  terrain->width * terrain->height);
}

void
t_terrain_invert (TTerrain *terrain)
{
  gint i;
  gint size;

  size = terrain->width * terrain->height;
  for (i = 0; i < size; i++)
    terrain->heightfield[i] = 1.0 - terrain->heightfield[i];
}

void
t_terrain_paint (TTerrain *terrain,
                 TRgb     *rgb)
{
  guchar *colormap;

  colormap = colormap_new (terrain->colormap, terrain->sealevel);

  switch (terrain->view)
    {
      case T_VIEW_2D_PLANE: 
	      t_terrain_paint_2d_plane (terrain, rgb, colormap); 
	      break;
      case T_VIEW_2D_CONTOUR: 
	      t_terrain_paint_2d_contour (terrain, rgb, colormap); 
	      break;
      case T_VIEW_3D_WIRE: 
	      t_terrain_paint_3d_wire (terrain, rgb, terrain->angle, 
	        terrain->elevation, terrain->wireframe_resolution, colormap); 
	      break;
      case T_VIEW_3D_HEIGHT: 
	      t_terrain_paint_3d_height (terrain, rgb, colormap); 
	      break;
      case T_VIEW_3D_LIGHT: 
	      t_terrain_paint_3d_light (terrain, rgb, colormap); 
	      break;
    }

  g_free (colormap);
}

void
t_terrain_paint_preview (TTerrain  *terrain,
                         TRgb      *rgb)
{
  guchar *colormap;

  colormap = colormap_new (T_COLORMAP_LAND, terrain->sealevel);
  t_terrain_paint_2d_plane (terrain, rgb, colormap);
  g_free (colormap);
}

void
t_terrain_paint_2d_plane (TTerrain *terrain,
                          TRgb     *rgb,
                          guchar   *palette)
{
  gint    x, y;
  gint    count;
  guchar *raster;

  g_return_if_fail (palette != NULL);

  raster = g_new (guchar, rgb->width * 3);

  for (y = 0; y < rgb->height; y++)
    {
      gint hf_pos;

      hf_pos = y * terrain->height / rgb->height * terrain->width;

      count = 0;
      for (x = 0; x < rgb->width * 3; x += 3)
        {
          gint value;

          value = terrain->heightfield[hf_pos] * 256;
          value = MAX (MIN (value, 255), 0) * 3;

          if (terrain->selection != NULL)
            {
              gint select;

              select = 255 - terrain->selection[hf_pos] * 127.0;

              raster[x + 0] = 255 - (((255 - palette[value + 0]) * select) >> 8);
              raster[x + 1] = 255 - (((255 - palette[value + 1]) * select) >> 8);
              raster[x + 2] = 255 - (((255 - palette[value + 2]) * select) >> 8);
            }
          else
            {
              raster[x + 0] = palette[value + 0];
              raster[x + 1] = palette[value + 1];
              raster[x + 2] = palette[value + 2];
            }

          count += terrain->width;
          while (count >= rgb->width)
            {
              count -= rgb->width;
              hf_pos++;
            }
        }

      t_rgb_draw_raster (rgb, y, raster);
    }

  g_free (raster);
}

void
t_terrain_paint_2d_contour (TTerrain *terrain,
                            TRgb     *rgb,
                            guchar   *palette)
{
  GList  *pos;
  gint    i, x, y, x2, y2;
  gfloat  scale;
  gint    count;
  guchar *raster;

  g_return_if_fail (palette != NULL);

  raster = g_new (guchar, rgb->width * 3);

  for (y = 0; y < rgb->height; y++)
    {
      gint hf_pos;

      hf_pos = y * terrain->height / rgb->height * terrain->width;

      count = 0;
      for (x = 0; x < rgb->width * 3; x += 3)
        {
          gint value;

          value = terrain->heightfield[hf_pos] * 256;
          value = MAX (MIN (value, 255), 0) * 3;

          raster[x + 0] = palette[value + 0];
          raster[x + 1] = palette[value + 1];
          raster[x + 2] = palette[value + 2];

          count += terrain->width;
          while (count >= rgb->width)
            {
              count -= rgb->width;
              hf_pos++;
            }
        }

      t_rgb_draw_raster (rgb, y, raster);
    }

  g_free (raster);

  scale = ((gfloat) rgb->width) / terrain->width;

  t_rgb_set_fore (rgb, 0, 0, 0);
  pos = g_list_first (terrain->contour_lines);
  while (pos != NULL)
    {
      GArray *points;

      points = pos->data;
      x = g_array_index (points, gfloat, 1) * scale;
      y = g_array_index (points, gfloat, 2) * scale;
      for (i = 3; i < points->len; i += 2)
        {
          x2 = g_array_index (points, gfloat, i + 0) * scale;
          y2 = g_array_index (points, gfloat, i + 1) * scale;
          t_rgb_draw_line (rgb, x, y, x2, y2);
          x = x2;
          y = y2;
        }

      pos = pos->next;
    }
}

void
t_terrain_paint_3d_wire (TTerrain *terrain,
                         TRgb     *rgb,
                         gfloat    angle,
                         gfloat    elevation /* -2.0 - 2.0 */,
                         gint      resolution,
                         guchar   *palette)
{
  gint   y;
  gint   division;
  gint   width_offset, height_offset;
  gfloat scrn_offset_x, scrn_offset_y;
  gfloat y_scale;
  gfloat terrain_size, rgb_size;
  gfloat cos_angle, sin_angle;
  gfloat camera_y, camera_z;
  gfloat zoom;
  gfloat half_width, half_height;

  division = (gint) ceil (((gfloat) terrain->width) / rgb->width * resolution);
  terrain_size = MAX (terrain->width, terrain->height);
  rgb_size = MAX (rgb->width, rgb->height);
  y_scale = terrain->y_scale_factor * terrain_size;

  t_rgb_clear (rgb);

  /* Add 0.5 to the elevation so that an elevation parameter of 0.0 centers
     the camera at the midpoint of a normalized heightfield. */
  camera_y = terrain_size * (elevation + 0.5);
  camera_z = -terrain_size * 1.5;
  zoom = rgb_size * 1.3;

  /* Use the offsets to center the plot onscreen. */
  scrn_offset_x = rgb->width * 0.5;
  scrn_offset_y =
    rgb->height * 0.5 + zoom * (camera_y - 0.5 * y_scale) / camera_z;

  cos_angle = cos (angle);
  sin_angle = sin (angle);

  width_offset = (terrain->width % division) >> 1;
  height_offset = (terrain->height % division) >> 1;
  half_width = terrain->width * 0.5;
  half_height = terrain->height * 0.5;

  for (y = height_offset; y < terrain->height; y += division)
    {
      gint x;
      gint pos;

      pos = y * terrain->width + width_offset;
      for (x = width_offset; x < terrain->width; x += division)
        {
          gfloat   temp_x, temp_y, temp_z;
          gfloat   coord_x, coord_y;
          gint     scrn_x_1, scrn_y_1;
          gint     scrn_x_2=0, scrn_y_2=0; /* silence compiler warnings */
          gint     scrn_x_3, scrn_y_3;
          gint     value;
          gboolean right;

          value = ((terrain->heightfield[pos] * 255.0) + 0.5);
          value = MAX (MIN (value, 255), 0) * 3;
 
          coord_x = x - half_width; coord_y = y - half_height;
          temp_x = cos_angle * coord_x + sin_angle * coord_y;
          temp_y = terrain->heightfield[pos];
          if (terrain->filled_sea && temp_y < terrain->sealevel)
            temp_y = terrain->sealevel;
          temp_y *= y_scale;
          temp_z = sin_angle * coord_x - cos_angle * coord_y;
          scrn_x_1 = rint (scrn_offset_x + zoom * temp_x / (temp_z - camera_z));
          scrn_y_1 = rint (scrn_offset_y - zoom * (temp_y - camera_y) / (temp_z - camera_z));

          t_rgb_set_fore (rgb, palette[value],
                          palette[value + 1], palette[value + 2]);

          right = FALSE;
          if (x + division < terrain->width)
            {
              coord_x = x + division - half_width; coord_y = y - half_height;
              temp_x = cos_angle * coord_x + sin_angle * coord_y;
              temp_y = terrain->heightfield[pos + division];
              if (terrain->filled_sea && temp_y < terrain->sealevel)
                temp_y = terrain->sealevel;
              temp_y *= y_scale;
              temp_z = sin_angle * coord_x - cos_angle * coord_y;
              scrn_x_2 = rint (scrn_offset_x + zoom * temp_x / (temp_z - camera_z));
              scrn_y_2 = rint (scrn_offset_y - zoom * (temp_y - camera_y) / (temp_z - camera_z));

              right = TRUE;

              t_rgb_draw_line_add (rgb, scrn_x_1, scrn_y_1, scrn_x_2, scrn_y_2);
            }

          if (y + division < terrain->height)
            {
              coord_x = x - half_width; coord_y = y + division - half_height;
              temp_x = cos_angle * coord_x + sin_angle * coord_y;
              temp_y = terrain->heightfield[pos + division * terrain->width];
              if (terrain->filled_sea && temp_y < terrain->sealevel)
                temp_y = terrain->sealevel;
              temp_y *= y_scale;
              temp_z = sin_angle * coord_x - cos_angle * coord_y;
              scrn_x_3 = rint (scrn_offset_x + zoom * temp_x / (temp_z - camera_z));
              scrn_y_3 = rint (scrn_offset_y - zoom * (temp_y - camera_y) / (temp_z - camera_z));

              t_rgb_draw_line_add (rgb, scrn_x_1, scrn_y_1, scrn_x_3, scrn_y_3);

              if (right)
                t_rgb_draw_line_add (rgb, scrn_x_2, scrn_y_2, scrn_x_3, scrn_y_3);
            }

          pos += division;
        }
    }
}

void
t_terrain_paint_3d_light (TTerrain  *terrain,
                          TRgb      *rgb,
                          guchar    *palette)
{
  gint   y;
  gint   scrn_offset_x, scrn_offset_y;
  gfloat y_scale;
  gfloat terrain_size, rgb_size;
  gfloat camera_y, camera_z;
  gfloat zoom;
  gfloat half_width, half_height;

  terrain_size = MAX (terrain->width, terrain->height);
  rgb_size = MAX (rgb->width, rgb->height);
  y_scale = terrain->y_scale_factor * terrain_size;

  t_rgb_clear (rgb);
  camera_y = terrain_size * 1.5;
  camera_z = -terrain_size * 1.5;
  zoom = rgb_size * 1.3;

  half_width = terrain->width * 0.5;
  half_height = terrain->height * 0.5;

  /* Use the offsets to center the plot onscreen. */
  scrn_offset_x = rgb->width / 2;
  scrn_offset_y = -rgb->height / 8 +
    rgb->height / 2 + zoom * (camera_y - 0.5 * y_scale) / camera_z;

  for (y = 0; y + 1 < terrain->height; y++)
    {
      gint   x;
      gint   pos;
      gint   temp_x_1, temp_y_1;
      gint   temp_x_2, temp_y_2;
      gfloat coord_x_1, coord_y_1, coord_z_1;
      gfloat coord_x_2, coord_y_2, coord_z_2;

      pos = y * terrain->width;

      coord_x_1 = -half_width;
      coord_y_1 = terrain->heightfield[pos];
      if (terrain->filled_sea && coord_y_1 < terrain->sealevel)
        coord_y_1 = terrain->sealevel;
      coord_y_1 *= y_scale;
      coord_z_1 = half_height - y;
      temp_x_1 = zoom * coord_x_1 / (coord_z_1 - camera_z);
      temp_y_1 = zoom * (coord_y_1 - camera_y) / (coord_z_1 - camera_z);

      coord_x_2 = -half_width;
      coord_y_2 = terrain->heightfield[pos + terrain->width];
      if (terrain->filled_sea && coord_y_2 < terrain->sealevel)
        coord_y_2 = terrain->sealevel;
      coord_y_2 *= y_scale;
      coord_z_2 = half_height - y - 1;
      temp_x_2 = zoom * coord_x_2 / (coord_z_2 - camera_z);
      temp_y_2 = zoom * (coord_y_2 - camera_y) / (coord_z_2 - camera_z);

      for (x = 0; x + 1 < terrain->width; x++)
        {
          gint     temp_x_3, temp_y_3;
          gint     temp_x_4, temp_y_4;
          gfloat   coord_x_3, coord_y_3, coord_z_3;
          gfloat   coord_x_4, coord_y_4, coord_z_4;
          gfloat   vector_y_2, vector_y_3;
          gfloat   light_x, light_y, light_z;
          gfloat   magnitude;
          gint     value;
          gfloat   light;
          GdkPoint points[3];

          value = ((terrain->heightfield[pos] * 255.0) + 0.5);
          value = MAX (MIN (value, 255), 0) * 3;
 
          coord_x_3 = x + 1 - half_width;
          coord_y_3 = terrain->heightfield[pos + 1];
          if (terrain->filled_sea && coord_y_3 < terrain->sealevel)
            coord_y_3 = terrain->sealevel;
          coord_y_3 *= y_scale;
          coord_z_3 = half_height - y;
          temp_x_3 = zoom * coord_x_3 / (coord_z_3 - camera_z);
          temp_y_3 = zoom * (coord_y_3 - camera_y) / (coord_z_3 - camera_z);

          coord_x_4 = x + 1 - half_width;
          coord_y_4 = terrain->heightfield[pos + terrain->width + 1];
          if (terrain->filled_sea && coord_y_4 < terrain->sealevel)
            coord_y_4 = terrain->sealevel;
          coord_y_4 *= y_scale;
          coord_z_4 = half_height - y - 1;
          temp_x_4 = zoom * coord_x_4 / (coord_z_4 - camera_z);
          temp_y_4 = zoom * (coord_y_4 - camera_y) / (coord_z_4 - camera_z);

          /* Calculate the light for the upper left triangle.  We
             also draw the lower right triangle using the same value
             and lighting to save CPU time. */
          vector_y_2 = coord_y_2 - coord_y_1;
          magnitude = sqrt (1.0 + vector_y_2 * vector_y_2);
          vector_y_2 /= magnitude;

          vector_y_3 = coord_y_3 - coord_y_1;
          magnitude = sqrt (1.0 + vector_y_3 * vector_y_3);
          vector_y_3 /= magnitude;

          /* Dot product */
          magnitude = sqrt (vector_y_2 * vector_y_2 + 1.0 +
                            vector_y_3 * vector_y_3);
          coord_x_1 = vector_y_2 / magnitude;
          coord_y_1 = 1.0 / magnitude;
          coord_z_1 = vector_y_3 / magnitude;

          light_x = 0.5773;
          light_y = 0.5773;
          light_z = 0.5773;

          light = 1.0 - pow ((coord_x_1 * light_x +
                              coord_y_1 * light_y +
                              coord_z_1 * light_z + 1.0) *
                             0.5, 2.0) * 0.5;

          t_rgb_set_fore (rgb, 255 - light * (255 - (palette[value] >> 1)),
                          255 - light * (255 - (palette[value + 1] >> 1)),
                          255 - light * (255 - (palette[value + 2] >> 1)));

          points[0].x = scrn_offset_x + temp_x_1;
          points[0].y = scrn_offset_y - temp_y_1;
          points[1].x = scrn_offset_x + temp_x_2;
          points[1].y = scrn_offset_y - temp_y_2;
          points[2].x = scrn_offset_x + temp_x_3;
          points[2].y = scrn_offset_y - temp_y_3;
          t_rgb_fill_triangle (rgb, points);

          points[0].x = scrn_offset_x + temp_x_4;
          points[0].y = scrn_offset_y - temp_y_4;
          t_rgb_fill_triangle (rgb, points);

          temp_x_1 = temp_x_3;
          temp_y_1 = temp_y_3;
          temp_x_2 = temp_x_4;
          temp_y_2 = temp_y_4;
          coord_x_1 = coord_x_3;
          coord_y_1 = coord_y_3;
          coord_x_2 = coord_x_4;
          coord_y_2 = coord_y_4;

          pos++;
        }
    }
}

void
t_terrain_paint_3d_height (TTerrain  *terrain,
                           TRgb      *rgb,
                           guchar    *palette)
{
  gint   y;
  gint   scrn_offset_x, scrn_offset_y;
  gfloat y_scale;
  gfloat terrain_size, rgb_size;
  gfloat camera_y, camera_z;
  gfloat zoom;
  gfloat half_width, half_height;

  terrain_size = MAX (terrain->width, terrain->height);
  rgb_size = MAX (rgb->width, rgb->height);
  y_scale = terrain->y_scale_factor * terrain_size;

  t_rgb_clear (rgb);
  camera_y = terrain_size * 1.5;
  camera_z = -terrain_size * 1.5;
  zoom = rgb_size * 1.3;

  half_width = terrain->width * 0.5;
  half_height = terrain->height * 0.5;

  /* Use the offsets to center the plot onscreen. */
  scrn_offset_x = rgb->width / 2;
  scrn_offset_y = -rgb->height / 8 +
    rgb->height / 2 + zoom * (camera_y - 0.5 * y_scale) / camera_z;

  for (y = 0; y + 1 < terrain->height; y++)
    {
      gint x;
      gint pos;
      gfloat coord_x_1, coord_y_1, coord_z_1, coord_y_2;
      gint   temp_x_1, temp_y_1;
      gint   temp_x_2, temp_y_2;

      pos = y * terrain->width;

      coord_x_1 = -half_width;
      coord_y_1 = terrain->heightfield[pos];
      if (terrain->filled_sea && coord_y_1 < terrain->sealevel)
        coord_y_1 = terrain->sealevel;
      coord_y_1 *= y_scale;
      coord_z_1 = half_height - y;
      temp_x_1 = zoom * coord_x_1 / (coord_z_1 - camera_z);
      temp_y_1 = zoom * (coord_y_1 - camera_y) / (coord_z_1 - camera_z);

      coord_y_2 = terrain->heightfield[pos + terrain->width];
      if (terrain->filled_sea && coord_y_2 < terrain->sealevel)
        coord_y_2 = terrain->sealevel;
      coord_y_2 *= y_scale;
      temp_x_2 = zoom * coord_x_1 / ((coord_z_1 - 1) - camera_z);
      temp_y_2 = zoom * (coord_y_2 - camera_y) / ((coord_z_1 - 1) - camera_z);

      for (x = 0; x + 1 < terrain->width; x++)
        {
          gint     temp_x_3, temp_y_3;
          gint     temp_x_4, temp_y_4;
          gfloat   coord_y_3, coord_y_4;
          gint     value;
          GdkPoint pts[3];

          coord_x_1 = x - half_width;

          coord_y_3 = terrain->heightfield[pos + 1];
          if (terrain->filled_sea && coord_y_3 < terrain->sealevel)
            coord_y_3 = terrain->sealevel;
          coord_y_3 *= y_scale;
          temp_x_3 = zoom * (coord_x_1 + 1) / (coord_z_1 - camera_z);
          temp_y_3 = zoom * (coord_y_3 - camera_y) / (coord_z_1 - camera_z);

          coord_y_4 = terrain->heightfield[pos + terrain->width + 1];
          if (terrain->filled_sea && coord_y_4 < terrain->sealevel)
            coord_y_4 = terrain->sealevel;
          coord_y_4 *= y_scale;
          temp_x_4 = zoom * (coord_x_1 + 1) / ((coord_z_1 - 1) - camera_z);
          temp_y_4 = zoom * (coord_y_4 - camera_y) / ((coord_z_1 - 1) - camera_z);

          value = ((terrain->heightfield[pos] * 255.0) + 0.5);
          value = MAX (MIN (value, 255), 0) * 3;

          t_rgb_set_fore (rgb, palette[value],
                          palette[value + 1], palette[value + 2]);

          pts[0].x = scrn_offset_x + temp_x_1;
          pts[0].y = scrn_offset_y - temp_y_1;
          pts[1].x = scrn_offset_x + temp_x_2;
          pts[1].y = scrn_offset_y - temp_y_2;
          pts[2].x = scrn_offset_x + temp_x_3;
          pts[2].y = scrn_offset_y - temp_y_3;

          t_rgb_fill_triangle (rgb, pts);

          pts[0].x = scrn_offset_x + temp_x_4;
          pts[0].y = scrn_offset_y - temp_y_4;

          t_rgb_fill_triangle (rgb, pts);

          temp_x_1 = temp_x_3;
          temp_y_1 = temp_y_3;
          temp_x_2 = temp_x_4;
          temp_y_2 = temp_y_4;

          pos++;
        }
    }
}

void
t_terrain_pack_options (TTerrain  *terrain,
                        GtkWidget *options)
{
  GList *themes;

  themes = gtk_object_get_data (GTK_OBJECT (options), "data_themes");

  terrain->modified = TRUE;
  terrain->camera_x = get_float (options, "render_camera_x");
  terrain->camera_y = get_float (options, "render_camera_y");
  terrain->camera_z = get_float (options, "render_camera_z");
  terrain->lookat_x = get_float (options, "render_lookat_x");
  terrain->lookat_y = get_float (options, "render_lookat_y");
  terrain->lookat_z = get_float (options, "render_lookat_z");
  terrain->observe_sealevel = get_boolean (options, "render_observe_sealevel");
  terrain->elevation_offset = get_float (options, "render_elevation");
  terrain->smooth_terrain = get_boolean (options, "render_smooth");
  terrain->clouds = get_boolean (options, "render_clouds");
  terrain->fog = get_boolean (options, "render_fog");
  terrain->filled_sea = get_boolean (options, "render_filled_sea");
  terrain->time_of_day = get_float (options, "render_time");
  terrain->north_direction = get_float (options, "render_north");
  terrain->water_clarity = get_float (options, "render_clarity");
  terrain->render_width = get_int (options, "render_width");
  terrain->y_scale_factor = get_float (options, "render_y_scale_factor");
  terrain->sealevel = get_float (options, "render_sealevel");
  terrain->wireframe_resolution = get_int (options, "render_wireframe_resolution");
  g_free (terrain->theme_file);
  terrain->theme_file = g_strdup (
    g_list_nth_data (themes, get_option (options, "render_theme")));
  terrain->lighting_level = get_option (options, "render_lighting");
  terrain->contour_levels = get_int (options, "render_levels");
}

void
t_terrain_unpack_options (TTerrain  *terrain,
                          GtkWidget *options)
{
  GList *themes;
  gint   value;

  options = lookup_toplevel (options);
  themes = gtk_object_get_data (GTK_OBJECT (options), "data_themes");

  set_float (options, "render_camera_x", terrain->camera_x);
  set_float (options, "render_camera_y", terrain->camera_y);
  set_float (options, "render_camera_z", terrain->camera_z);
  set_float (options, "render_lookat_x", terrain->lookat_x);
  set_float (options, "render_lookat_y", terrain->lookat_y);
  set_float (options, "render_lookat_z", terrain->lookat_z);
  set_boolean (options, "render_observe_sealevel", terrain->observe_sealevel);
  set_float (options, "render_elevation", terrain->elevation_offset);
  set_boolean (options, "render_smooth", terrain->smooth_terrain);
  set_boolean (options, "render_clouds", terrain->clouds);
  set_boolean (options, "render_fog", terrain->fog);
  set_boolean (options, "render_filled_sea", terrain->filled_sea);
  set_float (options, "render_time", terrain->time_of_day);
  set_float (options, "render_north", terrain->north_direction);
  set_float (options, "render_clarity", terrain->water_clarity);
  set_float (options, "render_y_scale_factor", terrain->y_scale_factor);
  set_float (options, "render_sealevel", terrain->sealevel);
  set_int (options, "render_wireframe_resolution",
           terrain->wireframe_resolution);
  set_int (options, "render_width", terrain->render_width);
  value = g_list_string_index (themes, terrain->theme_file);
  set_option (options, "render_theme", value >= 0 ? value : 0);
  set_option (options, "render_lighting", terrain->lighting_level);
  set_int (options, "render_levels", terrain->contour_levels);
} 

static gchar *preview_view_radio[] = {
  "_2d_plane", "_2d_contour",
  "_3d_wire", "_3d_height",
  "_3d_light", NULL
};

static gchar *preview_colormap_radio[] = {
  "land", "desert",
  "grayscale", "heat", NULL
};

/*
 * These pack/unpack terrain to/from widget functions should
 * go in terrainwindow.c instead.
 */

void
t_terrain_pack_preview_menu (TTerrain  *terrain,
                             GtkWidget *menu)
{
  GtkWidget *filter_window;
  TRgb      *rgb;
  gint       view;
  gint       colormap;
  gboolean   refresh;

  refresh = FALSE;

  filter_window = gtk_object_get_data (GTK_OBJECT (menu), "data_parent");
  view = get_radio_menu (menu, preview_view_radio);

  if (view == T_VIEW_2D_CONTOUR && terrain->view != T_VIEW_2D_CONTOUR)
    t_terrain_contour_lines (terrain, terrain->contour_levels, 5);

  if (view != -1 && terrain->view != view)
    {
      terrain->view = view;
      refresh = TRUE;
    }

  colormap = get_radio_menu (menu, preview_colormap_radio);
  if (colormap != -1 && terrain->colormap != colormap)
    {
      terrain->colormap = colormap;
      refresh = TRUE;
    }

  rgb = gtk_object_get_data (GTK_OBJECT (filter_window), "data_t_rgb");
  g_return_if_fail (rgb != NULL);

  filter_window_update_preview (filter_window);
  t_rgb_soft_sync (rgb);
}

void
t_terrain_unpack_preview_menu (TTerrain  *terrain,
                               GtkWidget *menu)
{
  set_radio_menu (menu, preview_view_radio, terrain->view);
  set_radio_menu (menu, preview_colormap_radio, terrain->colormap);
}

void
t_terrain_print_contour_map (TTerrain  *terrain,
                             GnomePrintContext *context)
{
  GList   *list;
  gdouble  scale;
  gdouble  translate_x, translate_y;
  gdouble  page_width, page_height;
  gdouble  usable_width, usable_height;

  page_width = 8.5 * 72.0;
  page_height = 11.0 * 72.0;

  usable_width = 7.0 * 72.0;
  usable_height = 9.0 * 72.0;

  scale = MIN (usable_width / terrain->width, usable_height / terrain->height);

  translate_x = (page_width - scale * terrain->width) * 0.5;
  translate_y = page_height - (page_height - scale * terrain->height) * 0.5;

  gnome_print_translate (context, translate_x, translate_y);
  gnome_print_scale (context, scale, -scale);

  gnome_print_moveto (context, 0.0, 0.0);
  gnome_print_lineto (context, terrain->width, 0.0);
  gnome_print_lineto (context, terrain->width, terrain->height);
  gnome_print_lineto (context, 0.0, terrain->height);
  gnome_print_lineto (context, 0.0, 0.0);
  gnome_print_stroke (context);

  t_terrain_contour_lines (terrain, terrain->contour_levels, 5);
  list = g_list_first (terrain->contour_lines);
  while (list != NULL)
    {
      GArray  *array;
      gdouble  x, y;
      gint     i;

      array = list->data;
      x = g_array_index (array, gfloat, 1);
      y = g_array_index (array, gfloat, 2);
      gnome_print_moveto (context, x, y);
      for (i = 1; i < array->len; i += 2)
        {
          x = g_array_index (array, gfloat, i + 0);
          y = g_array_index (array, gfloat, i + 1);
          gnome_print_lineto (context, x, y);
        }
      gnome_print_stroke (context);

      list = list->next;
    }
}

int
t_terrain_ref (TTerrain *terrain)
{
  return ++terrain->ref_count;
}

int
t_terrain_unref (TTerrain *terrain)
{
  if (terrain == NULL)
    return 0;

  if (--terrain->ref_count == 0)
    {
      t_terrain_free (terrain);

      return 0;
    }

  return terrain->ref_count;
}



TTerrain *
t_terrain_import_heightfield (gchar *filename)
{
  return NULL;
}

void
t_terrain_export_heightfield (TTerrain *terrain,
                              gchar    *filename)
{
}

static void
raster_set (TTerrain   *terrain,
            gint        x1,
            gint        y,
            gint        x2,
            TComposeOp  op)
{
  gint pos;

  if (y < 0 || y >= terrain->height ||
      x2 < 0 || x1 >= terrain->width ||
      x2 < x1)
    return;

  x1 = MAX (x1, 0);
  x2 = MIN (x2, terrain->width - 1);

  pos = y * terrain->width + x1;

  switch (op)
    {
    case T_COMPOSE_REPLACE:
    case T_COMPOSE_ADD:
      for (; x1 <= x2; x1++)
        terrain->selection[pos++] = 1.0;
      break;

    case T_COMPOSE_SUBTRACT:
      for (; x1 <= x2; x1++)
        terrain->selection[pos++] = 0.0;
      break;

    default:
      break;
    }
}

void
t_terrain_select (TTerrain       *terrain,
                  gfloat          x1,
                  gfloat          y1,
                  gfloat          x2,
                  gfloat          y2,
                  TSelectionType  type,
                  TComposeOp      op)
{
  gint y, _x1, _y1, _x2, _y2;

  _x1 = x1 * (terrain->width - 1);
  _y1 = y1 * (terrain->height - 1);
  _x2 = x2 * (terrain->width - 1);
  _y2 = y2 * (terrain->height - 1);

  if (op == T_COMPOSE_REPLACE)
    {
      if (type == T_SELECTION_NONE)
        {
          g_free (terrain->selection);
          terrain->selection = NULL;

          return;
        }
    }

  if (terrain->selection == NULL)
    terrain->selection = g_new0 (gfloat, terrain->width * terrain->height);
  else if (op == T_COMPOSE_REPLACE)
    memset (terrain->selection, 0,
            sizeof (gfloat) * terrain->width * terrain->height);

  switch (type)
    {
    case T_SELECTION_RECTANGLE:
      for (y = _y1; y <= _y2; y++)
        raster_set (terrain, _x1, y, _x2, op);
      break;

    case T_SELECTION_ELLIPSE:
      for (y = _y1; y <= _y2; y++)
        {
          gfloat value;
          gfloat x_radius;
          gint   x_start, x_stop;
          gfloat y_value;

          if (y < ((_y1 + _y2) >> 1))
            value = y - _y1;
          else
            value = _y2 - y;

          y_value = 1.0 - 2.0 * value / (_y2 - _y1 + 1);
          x_radius =
            sqrt (1.0 - y_value * y_value) * (_x2 - _x1 + 1) * 0.5;
          x_start = (_x1 + _x2) * 0.5 - x_radius;
          x_stop = (_x1 + _x2) * 0.5 + x_radius;
          raster_set (terrain, x_start, y, x_stop, op);
        }
      break;

      default:
        break;
    }
}

gint
t_terrain_add_object (TTerrain   *terrain,
                      gfloat      x,
                      gfloat      y,
                      gfloat      angle,
                      gfloat      scale_x,
                      gfloat      scale_y,
                      gfloat      scale_z,
                      gchar      *object_name)
{
  TTerrainObject object;

  object.name = g_strdup (object_name);
  object.x = x;
  object.y = y;
  object.angle = angle;
  object.scale_x = scale_x;
  object.scale_y = scale_y;
  object.scale_z = scale_z;

  g_array_append_val (terrain->objects, object);

  return terrain->objects->len - 1;
}

void
t_terrain_move_object (TTerrain   *terrain,
                       gint        item,
                       gfloat      x,
                       gfloat      y)
{
  TTerrainObject *object;

  object = &g_array_index (terrain->objects, TTerrainObject, item);
  object->x = x;
  object->y = y;
}

void
t_terrain_rotate_object (TTerrain   *terrain,
                         gint        item,
                         gfloat      angle)
{
  g_array_index (terrain->objects, TTerrainObject, item).angle = angle;
}

void
t_terrain_remove_objects (TTerrain   *terrain)
{
  gint i, size;

  size = terrain->objects->len;
  for (i = 0; i < size; i++)
    t_terrain_object_clear (&g_array_index (terrain->objects,
                            TTerrainObject, i));

  g_array_free (terrain->objects, TRUE);
  terrain->objects = g_array_new (FALSE, FALSE, sizeof (TTerrainObject));
}
