/* Terraform - (C) 1997-2000 Robert Gasch (r.gasch@chello.nl)
 *  - http://terraform.sourceforge.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#ifndef __TTERRAIN_H__
#define __TTERRAIN_H__

#include <gtk/gtk.h>
#include <libgnomeprint/gnome-print.h>
#include "trgb.h"
#include "colormaps.h"

typedef enum TComposeOp TComposeOp;

enum TComposeOp
{
  T_COMPOSE_NONE,
  T_COMPOSE_REPLACE,
  T_COMPOSE_ADD,
  T_COMPOSE_SUBTRACT
};

typedef enum TViewMd TViewMd;

enum TViewMd
{
  T_VIEW_2D_PLANE,
  T_VIEW_2D_CONTOUR,
  T_VIEW_3D_WIRE,
  T_VIEW_3D_HEIGHT,
  T_VIEW_3D_LIGHT
};

typedef struct TTerrainObject TTerrainObject;

struct TTerrainObject
{
  gfloat    x, y;
  gfloat    angle;
  gfloat    scale_x, scale_y, scale_z;
  gchar    *name;
};

void t_terrain_object_clear (TTerrainObject *object);

typedef struct TTerrain TTerrain;

struct TTerrain
{
  gint         ref_count;

  gchar       *filename;
  gboolean     modified;
  gboolean     autogenned; /* Set true if filename is auto generated. */

  gint         width, height;
  gfloat      *heightfield;
  gfloat      *selection;

  gfloat       camera_x, camera_y, camera_z;
  gfloat       lookat_x, lookat_y, lookat_z;
  gboolean     observe_sealevel;
  gfloat       elevation_offset;
  gboolean     smooth_terrain;
  gboolean     clouds;
  gboolean     fog;
  gboolean     filled_sea;
  gfloat       time_of_day;
  gfloat       north_direction;
  gfloat       water_clarity;
  gint         render_width;
  gfloat       y_scale_factor;
  gfloat       sealevel;
  gint         wireframe_resolution;
  gchar       *theme_file;
  gint         lighting_level;

  gint         contour_levels;

  TViewMd      view;
  TColormapTp colormap;
  gfloat      angle;
  gfloat      elevation;
  gboolean    auto_rotate;

  /* objects is a GArray of TTerrainObjects */
  GArray     *objects;

  /* contour_lines is a GList of GArrays of a gfloat height + gfloat pairs */
  GList      *contour_lines;
};


TTerrain *t_terrain_new                (gint        width,
                                        gint        height);
void      t_terrain_free               (TTerrain   *terrain);
TTerrain *t_terrain_clone              (TTerrain   *terrain);
TTerrain *t_terrain_clone_resize       (TTerrain   *terrain,
                                        gint        width,
                                        gint        height);
void      t_terrain_set_size           (TTerrain   *terrain,
                                        gint        width,
                                        gint        height);
TTerrain *t_terrain_clone_preview      (TTerrain   *terrain);
void      t_terrain_copy_heightfield   (TTerrain   *from,
                                        TTerrain   *to);
void      t_terrain_copy_heightfield_and_selection (TTerrain   *from,
                                                    TTerrain   *to);
void      t_terrain_normalize          (TTerrain   *terrain,
                                        gboolean    never_grow);
void      t_terrain_crop               (TTerrain   *terrain,
                                        gint        x1,
                                        gint        y1,
                                        gint        x2,
                                        gint        y2);
void      t_terrain_invert             (TTerrain   *terrain);

void      t_terrain_paint              (TTerrain   *terrain,
                                        TRgb       *rgb);
void      t_terrain_paint_preview      (TTerrain   *terrain,
                                        TRgb       *rgb);

void      t_terrain_paint_2d_plane     (TTerrain   *terrain,
                                        TRgb       *rgb,
                                        guchar     *palette);

void      t_terrain_paint_2d_contour   (TTerrain   *terrain,
                                        TRgb       *rgb,
                                        guchar     *palette);

void      t_terrain_paint_3d_wire      (TTerrain   *terrain,
                                        TRgb       *rgb,
                                        gfloat      angle,
                                        gfloat      elevation,
                                        gint        division,
                                        guchar     *palette);

void      t_terrain_paint_3d_light     (TTerrain   *terrain,
                                        TRgb       *rgb,
                                        guchar     *palette);

void      t_terrain_paint_3d_height    (TTerrain   *terrain,
                                        TRgb       *rgb,
                                        guchar     *palette);

void      t_terrain_pack_options       (TTerrain   *terrain,
                                        GtkWidget  *options);
void      t_terrain_unpack_options     (TTerrain   *terrain,
                                        GtkWidget  *options);
void      t_terrain_pack_preview_menu  (TTerrain   *terrain,
                                        GtkWidget  *menu);
void      t_terrain_unpack_preview_menu(TTerrain   *terrain,
                                        GtkWidget  *menu);
void      t_terrain_print_contour_map  (TTerrain   *terrain,
                                        GnomePrintContext *context);

int       t_terrain_ref                (TTerrain   *terrain);
int       t_terrain_unref              (TTerrain   *terrain);

TTerrain *t_terrain_import_heightfield (gchar      *filename);
void      t_terrain_export_heightfield (TTerrain   *terrain,
                                        gchar      *filename);
void      t_terrain_select             (TTerrain   *terrain,
                                        gfloat      x1,
                                        gfloat      y1,
                                        gfloat      x2,
                                        gfloat      y2,
                                        TSelectionType type,
                                        TComposeOp  op);
gint      t_terrain_add_object         (TTerrain   *terrain,
                                        gfloat      x,
                                        gfloat      y,
                                        gfloat      angle,
                                        gfloat      scale_x,
                                        gfloat      scale_y,
                                        gfloat      scale_z,
                                        gchar      *object_name);
void      t_terrain_move_object        (TTerrain   *terrain,
                                        gint        item,
                                        gfloat      x,
                                        gfloat      y);
void      t_terrain_rotate_object      (TTerrain   *terrain,
                                        gint        item,
                                        gfloat      angle);
void      t_terrain_remove_objects     (TTerrain   *terrain);

#endif /* __TTERRAIN_H__ */
